import sys
from datetime import datetime
from typing import Dict, List, Optional

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from localstack.aws.api import RequestContext, ServiceException, ServiceRequest, handler

ARN = str
BackendRole = str
Boolean = bool
CloudWatchLogsLogGroupArn = str
CommitMessage = str
ConnectionAlias = str
ConnectionId = str
ConnectionStatusMessage = str
DeploymentType = str
DescribePackagesFilterValue = str
DomainId = str
DomainName = str
DomainNameFqdn = str
Double = float
DryRun = bool
ErrorMessage = str
ErrorType = str
GUID = str
IdentityPoolId = str
InstanceCount = int
InstanceRole = str
Integer = int
IntegerClass = int
Issue = str
KmsKeyId = str
LimitName = str
LimitValue = str
MaxResults = int
MaximumInstanceCount = int
Message = str
MinimumInstanceCount = int
NextToken = str
NonEmptyString = str
OwnerId = str
PackageDescription = str
PackageID = str
PackageName = str
PackageVersion = str
Password = str
PolicyDocument = str
ReferencePath = str
Region = str
ReservationToken = str
RoleArn = str
S3BucketName = str
S3Key = str
SAMLEntityId = str
SAMLMetadata = str
ScheduledAutoTuneDescription = str
ServiceUrl = str
StorageSubTypeName = str
StorageTypeName = str
String = str
TagKey = str
TagValue = str
UIntValue = int
UpgradeName = str
UserPoolId = str
Username = str
VersionString = str


class AutoTuneDesiredState(str):
    ENABLED = "ENABLED"
    DISABLED = "DISABLED"


class AutoTuneState(str):
    ENABLED = "ENABLED"
    DISABLED = "DISABLED"
    ENABLE_IN_PROGRESS = "ENABLE_IN_PROGRESS"
    DISABLE_IN_PROGRESS = "DISABLE_IN_PROGRESS"
    DISABLED_AND_ROLLBACK_SCHEDULED = "DISABLED_AND_ROLLBACK_SCHEDULED"
    DISABLED_AND_ROLLBACK_IN_PROGRESS = "DISABLED_AND_ROLLBACK_IN_PROGRESS"
    DISABLED_AND_ROLLBACK_COMPLETE = "DISABLED_AND_ROLLBACK_COMPLETE"
    DISABLED_AND_ROLLBACK_ERROR = "DISABLED_AND_ROLLBACK_ERROR"
    ERROR = "ERROR"


class AutoTuneType(str):
    SCHEDULED_ACTION = "SCHEDULED_ACTION"


class DeploymentStatus(str):
    PENDING_UPDATE = "PENDING_UPDATE"
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"
    NOT_ELIGIBLE = "NOT_ELIGIBLE"
    ELIGIBLE = "ELIGIBLE"


class DescribePackagesFilterName(str):
    PackageID = "PackageID"
    PackageName = "PackageName"
    PackageStatus = "PackageStatus"


class DomainPackageStatus(str):
    ASSOCIATING = "ASSOCIATING"
    ASSOCIATION_FAILED = "ASSOCIATION_FAILED"
    ACTIVE = "ACTIVE"
    DISSOCIATING = "DISSOCIATING"
    DISSOCIATION_FAILED = "DISSOCIATION_FAILED"


class EngineType(str):
    OpenSearch = "OpenSearch"
    Elasticsearch = "Elasticsearch"


class InboundConnectionStatusCode(str):
    PENDING_ACCEPTANCE = "PENDING_ACCEPTANCE"
    APPROVED = "APPROVED"
    PROVISIONING = "PROVISIONING"
    ACTIVE = "ACTIVE"
    REJECTING = "REJECTING"
    REJECTED = "REJECTED"
    DELETING = "DELETING"
    DELETED = "DELETED"


class LogType(str):
    INDEX_SLOW_LOGS = "INDEX_SLOW_LOGS"
    SEARCH_SLOW_LOGS = "SEARCH_SLOW_LOGS"
    ES_APPLICATION_LOGS = "ES_APPLICATION_LOGS"
    AUDIT_LOGS = "AUDIT_LOGS"


class OpenSearchPartitionInstanceType(str):
    m3_medium_search = "m3.medium.search"
    m3_large_search = "m3.large.search"
    m3_xlarge_search = "m3.xlarge.search"
    m3_2xlarge_search = "m3.2xlarge.search"
    m4_large_search = "m4.large.search"
    m4_xlarge_search = "m4.xlarge.search"
    m4_2xlarge_search = "m4.2xlarge.search"
    m4_4xlarge_search = "m4.4xlarge.search"
    m4_10xlarge_search = "m4.10xlarge.search"
    m5_large_search = "m5.large.search"
    m5_xlarge_search = "m5.xlarge.search"
    m5_2xlarge_search = "m5.2xlarge.search"
    m5_4xlarge_search = "m5.4xlarge.search"
    m5_12xlarge_search = "m5.12xlarge.search"
    m5_24xlarge_search = "m5.24xlarge.search"
    r5_large_search = "r5.large.search"
    r5_xlarge_search = "r5.xlarge.search"
    r5_2xlarge_search = "r5.2xlarge.search"
    r5_4xlarge_search = "r5.4xlarge.search"
    r5_12xlarge_search = "r5.12xlarge.search"
    r5_24xlarge_search = "r5.24xlarge.search"
    c5_large_search = "c5.large.search"
    c5_xlarge_search = "c5.xlarge.search"
    c5_2xlarge_search = "c5.2xlarge.search"
    c5_4xlarge_search = "c5.4xlarge.search"
    c5_9xlarge_search = "c5.9xlarge.search"
    c5_18xlarge_search = "c5.18xlarge.search"
    t3_nano_search = "t3.nano.search"
    t3_micro_search = "t3.micro.search"
    t3_small_search = "t3.small.search"
    t3_medium_search = "t3.medium.search"
    t3_large_search = "t3.large.search"
    t3_xlarge_search = "t3.xlarge.search"
    t3_2xlarge_search = "t3.2xlarge.search"
    ultrawarm1_medium_search = "ultrawarm1.medium.search"
    ultrawarm1_large_search = "ultrawarm1.large.search"
    ultrawarm1_xlarge_search = "ultrawarm1.xlarge.search"
    t2_micro_search = "t2.micro.search"
    t2_small_search = "t2.small.search"
    t2_medium_search = "t2.medium.search"
    r3_large_search = "r3.large.search"
    r3_xlarge_search = "r3.xlarge.search"
    r3_2xlarge_search = "r3.2xlarge.search"
    r3_4xlarge_search = "r3.4xlarge.search"
    r3_8xlarge_search = "r3.8xlarge.search"
    i2_xlarge_search = "i2.xlarge.search"
    i2_2xlarge_search = "i2.2xlarge.search"
    d2_xlarge_search = "d2.xlarge.search"
    d2_2xlarge_search = "d2.2xlarge.search"
    d2_4xlarge_search = "d2.4xlarge.search"
    d2_8xlarge_search = "d2.8xlarge.search"
    c4_large_search = "c4.large.search"
    c4_xlarge_search = "c4.xlarge.search"
    c4_2xlarge_search = "c4.2xlarge.search"
    c4_4xlarge_search = "c4.4xlarge.search"
    c4_8xlarge_search = "c4.8xlarge.search"
    r4_large_search = "r4.large.search"
    r4_xlarge_search = "r4.xlarge.search"
    r4_2xlarge_search = "r4.2xlarge.search"
    r4_4xlarge_search = "r4.4xlarge.search"
    r4_8xlarge_search = "r4.8xlarge.search"
    r4_16xlarge_search = "r4.16xlarge.search"
    i3_large_search = "i3.large.search"
    i3_xlarge_search = "i3.xlarge.search"
    i3_2xlarge_search = "i3.2xlarge.search"
    i3_4xlarge_search = "i3.4xlarge.search"
    i3_8xlarge_search = "i3.8xlarge.search"
    i3_16xlarge_search = "i3.16xlarge.search"
    r6g_large_search = "r6g.large.search"
    r6g_xlarge_search = "r6g.xlarge.search"
    r6g_2xlarge_search = "r6g.2xlarge.search"
    r6g_4xlarge_search = "r6g.4xlarge.search"
    r6g_8xlarge_search = "r6g.8xlarge.search"
    r6g_12xlarge_search = "r6g.12xlarge.search"
    m6g_large_search = "m6g.large.search"
    m6g_xlarge_search = "m6g.xlarge.search"
    m6g_2xlarge_search = "m6g.2xlarge.search"
    m6g_4xlarge_search = "m6g.4xlarge.search"
    m6g_8xlarge_search = "m6g.8xlarge.search"
    m6g_12xlarge_search = "m6g.12xlarge.search"
    c6g_large_search = "c6g.large.search"
    c6g_xlarge_search = "c6g.xlarge.search"
    c6g_2xlarge_search = "c6g.2xlarge.search"
    c6g_4xlarge_search = "c6g.4xlarge.search"
    c6g_8xlarge_search = "c6g.8xlarge.search"
    c6g_12xlarge_search = "c6g.12xlarge.search"
    r6gd_large_search = "r6gd.large.search"
    r6gd_xlarge_search = "r6gd.xlarge.search"
    r6gd_2xlarge_search = "r6gd.2xlarge.search"
    r6gd_4xlarge_search = "r6gd.4xlarge.search"
    r6gd_8xlarge_search = "r6gd.8xlarge.search"
    r6gd_12xlarge_search = "r6gd.12xlarge.search"
    r6gd_16xlarge_search = "r6gd.16xlarge.search"
    t4g_small_search = "t4g.small.search"
    t4g_medium_search = "t4g.medium.search"


class OpenSearchWarmPartitionInstanceType(str):
    ultrawarm1_medium_search = "ultrawarm1.medium.search"
    ultrawarm1_large_search = "ultrawarm1.large.search"
    ultrawarm1_xlarge_search = "ultrawarm1.xlarge.search"


class OptionState(str):
    RequiresIndexDocuments = "RequiresIndexDocuments"
    Processing = "Processing"
    Active = "Active"


class OutboundConnectionStatusCode(str):
    VALIDATING = "VALIDATING"
    VALIDATION_FAILED = "VALIDATION_FAILED"
    PENDING_ACCEPTANCE = "PENDING_ACCEPTANCE"
    APPROVED = "APPROVED"
    PROVISIONING = "PROVISIONING"
    ACTIVE = "ACTIVE"
    REJECTING = "REJECTING"
    REJECTED = "REJECTED"
    DELETING = "DELETING"
    DELETED = "DELETED"


class PackageStatus(str):
    COPYING = "COPYING"
    COPY_FAILED = "COPY_FAILED"
    VALIDATING = "VALIDATING"
    VALIDATION_FAILED = "VALIDATION_FAILED"
    AVAILABLE = "AVAILABLE"
    DELETING = "DELETING"
    DELETED = "DELETED"
    DELETE_FAILED = "DELETE_FAILED"


class PackageType(str):
    TXT_DICTIONARY = "TXT-DICTIONARY"


class ReservedInstancePaymentOption(str):
    ALL_UPFRONT = "ALL_UPFRONT"
    PARTIAL_UPFRONT = "PARTIAL_UPFRONT"
    NO_UPFRONT = "NO_UPFRONT"


class RollbackOnDisable(str):
    NO_ROLLBACK = "NO_ROLLBACK"
    DEFAULT_ROLLBACK = "DEFAULT_ROLLBACK"


class ScheduledAutoTuneActionType(str):
    JVM_HEAP_SIZE_TUNING = "JVM_HEAP_SIZE_TUNING"
    JVM_YOUNG_GEN_TUNING = "JVM_YOUNG_GEN_TUNING"


class ScheduledAutoTuneSeverityType(str):
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"


class TLSSecurityPolicy(str):
    Policy_Min_TLS_1_0_2019_07 = "Policy-Min-TLS-1-0-2019-07"
    Policy_Min_TLS_1_2_2019_07 = "Policy-Min-TLS-1-2-2019-07"


class TimeUnit(str):
    HOURS = "HOURS"


class UpgradeStatus(str):
    IN_PROGRESS = "IN_PROGRESS"
    SUCCEEDED = "SUCCEEDED"
    SUCCEEDED_WITH_ISSUES = "SUCCEEDED_WITH_ISSUES"
    FAILED = "FAILED"


class UpgradeStep(str):
    PRE_UPGRADE_CHECK = "PRE_UPGRADE_CHECK"
    SNAPSHOT = "SNAPSHOT"
    UPGRADE = "UPGRADE"


class VolumeType(str):
    standard = "standard"
    gp2 = "gp2"
    io1 = "io1"


class AccessDeniedException(ServiceException):
    """An error occurred because user does not have permissions to access the
    resource. Returns HTTP status code 403.
    """

    pass


class BaseException(ServiceException):
    """An error occurred while processing the request."""

    message: Optional[ErrorMessage]


class ConflictException(ServiceException):
    """An error occurred because the client attempts to remove a resource that
    is currently in use. Returns HTTP status code 409.
    """

    pass


class DisabledOperationException(ServiceException):
    """An error occured because the client wanted to access a not supported
    operation. Gives http status code of 409.
    """

    pass


class InternalException(ServiceException):
    """The request processing has failed because of an unknown error, exception
    or failure (the failure is internal to the service) . Gives http status
    code of 500.
    """

    pass


class InvalidPaginationTokenException(ServiceException):
    """The request processing has failed because of invalid pagination token
    provided by customer. Returns an HTTP status code of 400.
    """

    pass


class InvalidTypeException(ServiceException):
    """An exception for trying to create or access sub-resource that is either
    invalid or not supported. Gives http status code of 409.
    """

    pass


class LimitExceededException(ServiceException):
    """An exception for trying to create more than allowed resources or
    sub-resources. Gives http status code of 409.
    """

    pass


class ResourceAlreadyExistsException(ServiceException):
    """An exception for creating a resource that already exists. Gives http
    status code of 400.
    """

    pass


class ResourceNotFoundException(ServiceException):
    """An exception for accessing or deleting a resource that does not exist.
    Gives http status code of 400.
    """

    pass


class ValidationException(ServiceException):
    """An exception for missing / invalid input fields. Gives http status code
    of 400.
    """

    pass


class AWSDomainInformation(TypedDict, total=False):
    OwnerId: Optional[OwnerId]
    DomainName: DomainName
    Region: Optional[Region]


class AcceptInboundConnectionRequest(ServiceRequest):
    """Container for the parameters to the ``AcceptInboundConnection``
    operation.
    """

    ConnectionId: ConnectionId


class InboundConnectionStatus(TypedDict, total=False):
    """The connection status of an inbound cross-cluster connection."""

    StatusCode: Optional[InboundConnectionStatusCode]
    Message: Optional[ConnectionStatusMessage]


class DomainInformationContainer(TypedDict, total=False):
    AWSDomainInformation: Optional[AWSDomainInformation]


class InboundConnection(TypedDict, total=False):
    """Details of an inbound connection."""

    LocalDomainInfo: Optional[DomainInformationContainer]
    RemoteDomainInfo: Optional[DomainInformationContainer]
    ConnectionId: Optional[ConnectionId]
    ConnectionStatus: Optional[InboundConnectionStatus]


class AcceptInboundConnectionResponse(TypedDict, total=False):
    """The result of an ``AcceptInboundConnection`` operation. Contains details
    about the accepted inbound connection.
    """

    Connection: Optional[InboundConnection]


UpdateTimestamp = datetime


class OptionStatus(TypedDict, total=False):
    """Provides the current status of the entity."""

    CreationDate: UpdateTimestamp
    UpdateDate: UpdateTimestamp
    UpdateVersion: Optional[UIntValue]
    State: OptionState
    PendingDeletion: Optional[Boolean]


class AccessPoliciesStatus(TypedDict, total=False):
    """The configured access rules for the domain's document and search
    endpoints, and the current status of those rules.
    """

    Options: PolicyDocument
    Status: OptionStatus


class Tag(TypedDict, total=False):
    """A key value pair for a resource tag."""

    Key: TagKey
    Value: TagValue


TagList = List[Tag]


class AddTagsRequest(ServiceRequest):
    """Container for the parameters to the ``AddTags`` operation. Specifies the
    tags to attach to the domain.
    """

    ARN: ARN
    TagList: TagList


LimitValueList = List[LimitValue]


class AdditionalLimit(TypedDict, total=False):
    """List of limits that are specific to a given InstanceType and for each of
    its ``InstanceRole`` .
    """

    LimitName: Optional[LimitName]
    LimitValues: Optional[LimitValueList]


AdditionalLimitList = List[AdditionalLimit]
AdvancedOptions = Dict[String, String]


class AdvancedOptionsStatus(TypedDict, total=False):
    """Status of the advanced options for the specified domain. Currently, the
    following advanced options are available:

    -  Option to allow references to indices in an HTTP request body. Must
       be ``false`` when configuring access to individual sub-resources. By
       default, the value is ``true``. See `Advanced cluster
       parameters <http://docs.aws.amazon.com/opensearch-service/latest/developerguide/createupdatedomains.html#createdomain-configure-advanced-options>`__
       for more information.
    -  Option to specify the percentage of heap space allocated to field
       data. By default, this setting is unbounded.

    For more information, see `Advanced cluster
    parameters <http://docs.aws.amazon.com/opensearch-service/latest/developerguide/createupdatedomains.html#createdomain-configure-advanced-options>`__.
    """

    Options: AdvancedOptions
    Status: OptionStatus


DisableTimestamp = datetime


class SAMLIdp(TypedDict, total=False):
    """The SAML identity povider's information."""

    MetadataContent: SAMLMetadata
    EntityId: SAMLEntityId


class SAMLOptionsOutput(TypedDict, total=False):
    """Describes the SAML application configured for the domain."""

    Enabled: Optional[Boolean]
    Idp: Optional[SAMLIdp]
    SubjectKey: Optional[String]
    RolesKey: Optional[String]
    SessionTimeoutMinutes: Optional[IntegerClass]


class AdvancedSecurityOptions(TypedDict, total=False):
    """The advanced security configuration: whether advanced security is
    enabled, whether the internal database option is enabled.
    """

    Enabled: Optional[Boolean]
    InternalUserDatabaseEnabled: Optional[Boolean]
    SAMLOptions: Optional[SAMLOptionsOutput]
    AnonymousAuthDisableDate: Optional[DisableTimestamp]
    AnonymousAuthEnabled: Optional[Boolean]


class SAMLOptionsInput(TypedDict, total=False):
    """The SAML application configuration for the domain."""

    Enabled: Optional[Boolean]
    Idp: Optional[SAMLIdp]
    MasterUserName: Optional[Username]
    MasterBackendRole: Optional[BackendRole]
    SubjectKey: Optional[String]
    RolesKey: Optional[String]
    SessionTimeoutMinutes: Optional[IntegerClass]


class MasterUserOptions(TypedDict, total=False):
    """Credentials for the master user: username and password, ARN, or both."""

    MasterUserARN: Optional[ARN]
    MasterUserName: Optional[Username]
    MasterUserPassword: Optional[Password]


class AdvancedSecurityOptionsInput(TypedDict, total=False):
    """The advanced security configuration: whether advanced security is
    enabled, whether the internal database option is enabled, master
    username and password (if internal database is enabled), and master user
    ARN (if IAM is enabled).
    """

    Enabled: Optional[Boolean]
    InternalUserDatabaseEnabled: Optional[Boolean]
    MasterUserOptions: Optional[MasterUserOptions]
    SAMLOptions: Optional[SAMLOptionsInput]
    AnonymousAuthEnabled: Optional[Boolean]


class AdvancedSecurityOptionsStatus(TypedDict, total=False):
    """The status of advanced security options for the specified domain."""

    Options: AdvancedSecurityOptions
    Status: OptionStatus


class AssociatePackageRequest(ServiceRequest):
    """Container for the request parameters to the ``AssociatePackage``
    operation.
    """

    PackageID: PackageID
    DomainName: DomainName


class ErrorDetails(TypedDict, total=False):
    ErrorType: Optional[ErrorType]
    ErrorMessage: Optional[ErrorMessage]


LastUpdated = datetime


class DomainPackageDetails(TypedDict, total=False):
    """Information on a package associated with a domain."""

    PackageID: Optional[PackageID]
    PackageName: Optional[PackageName]
    PackageType: Optional[PackageType]
    LastUpdated: Optional[LastUpdated]
    DomainName: Optional[DomainName]
    DomainPackageStatus: Optional[DomainPackageStatus]
    PackageVersion: Optional[PackageVersion]
    ReferencePath: Optional[ReferencePath]
    ErrorDetails: Optional[ErrorDetails]


class AssociatePackageResponse(TypedDict, total=False):
    """Container for the response returned by ``AssociatePackage`` operation."""

    DomainPackageDetails: Optional[DomainPackageDetails]


AutoTuneDate = datetime


class ScheduledAutoTuneDetails(TypedDict, total=False):
    """Specifies details about the scheduled Auto-Tune action. See `Auto-Tune
    for Amazon OpenSearch
    Service <https://docs.aws.amazon.com/opensearch-service/latest/developerguide/auto-tune.html>`__
    for more information.
    """

    Date: Optional[AutoTuneDate]
    ActionType: Optional[ScheduledAutoTuneActionType]
    Action: Optional[ScheduledAutoTuneDescription]
    Severity: Optional[ScheduledAutoTuneSeverityType]


class AutoTuneDetails(TypedDict, total=False):
    """Specifies details about the Auto-Tune action. See `Auto-Tune for Amazon
    OpenSearch
    Service <https://docs.aws.amazon.com/opensearch-service/latest/developerguide/auto-tune.html>`__
    for more information.
    """

    ScheduledAutoTuneDetails: Optional[ScheduledAutoTuneDetails]


class AutoTune(TypedDict, total=False):
    """Specifies the Auto-Tune type and Auto-Tune action details."""

    AutoTuneType: Optional[AutoTuneType]
    AutoTuneDetails: Optional[AutoTuneDetails]


AutoTuneList = List[AutoTune]
DurationValue = int


class Duration(TypedDict, total=False):
    """The maintenance schedule duration: duration value and duration unit. See
    `Auto-Tune for Amazon OpenSearch
    Service <https://docs.aws.amazon.com/opensearch-service/latest/developerguide/auto-tune.html>`__
    for more information.
    """

    Value: Optional[DurationValue]
    Unit: Optional[TimeUnit]


StartAt = datetime


class AutoTuneMaintenanceSchedule(TypedDict, total=False):
    """Specifies the Auto-Tune maintenance schedule. See `Auto-Tune for Amazon
    OpenSearch
    Service <https://docs.aws.amazon.com/opensearch-service/latest/developerguide/auto-tune.html>`__
    for more information.
    """

    StartAt: Optional[StartAt]
    Duration: Optional[Duration]
    CronExpressionForRecurrence: Optional[String]


AutoTuneMaintenanceScheduleList = List[AutoTuneMaintenanceSchedule]


class AutoTuneOptions(TypedDict, total=False):
    """The Auto-Tune options: the Auto-Tune desired state for the domain,
    rollback state when disabling Auto-Tune options and list of maintenance
    schedules.
    """

    DesiredState: Optional[AutoTuneDesiredState]
    RollbackOnDisable: Optional[RollbackOnDisable]
    MaintenanceSchedules: Optional[AutoTuneMaintenanceScheduleList]


class AutoTuneOptionsInput(TypedDict, total=False):
    """The Auto-Tune options: the Auto-Tune desired state for the domain and
    list of maintenance schedules.
    """

    DesiredState: Optional[AutoTuneDesiredState]
    MaintenanceSchedules: Optional[AutoTuneMaintenanceScheduleList]


class AutoTuneOptionsOutput(TypedDict, total=False):
    """The Auto-Tune options: the Auto-Tune desired state for the domain and
    list of maintenance schedules.
    """

    State: Optional[AutoTuneState]
    ErrorMessage: Optional[String]


class AutoTuneStatus(TypedDict, total=False):
    """Provides the current Auto-Tune status for the domain."""

    CreationDate: UpdateTimestamp
    UpdateDate: UpdateTimestamp
    UpdateVersion: Optional[UIntValue]
    State: AutoTuneState
    ErrorMessage: Optional[String]
    PendingDeletion: Optional[Boolean]


class AutoTuneOptionsStatus(TypedDict, total=False):
    """The Auto-Tune status for the domain."""

    Options: Optional[AutoTuneOptions]
    Status: Optional[AutoTuneStatus]


class CancelServiceSoftwareUpdateRequest(ServiceRequest):
    """Container for the parameters to the ``CancelServiceSoftwareUpdate``
    operation. Specifies the name of the domain that you wish to cancel a
    service software update on.
    """

    DomainName: DomainName


DeploymentCloseDateTimeStamp = datetime


class ServiceSoftwareOptions(TypedDict, total=False):
    """The current options of an domain service software options."""

    CurrentVersion: Optional[String]
    NewVersion: Optional[String]
    UpdateAvailable: Optional[Boolean]
    Cancellable: Optional[Boolean]
    UpdateStatus: Optional[DeploymentStatus]
    Description: Optional[String]
    AutomatedUpdateDate: Optional[DeploymentCloseDateTimeStamp]
    OptionalDeployment: Optional[Boolean]


class CancelServiceSoftwareUpdateResponse(TypedDict, total=False):
    """The result of a ``CancelServiceSoftwareUpdate`` operation. Contains the
    status of the update.
    """

    ServiceSoftwareOptions: Optional[ServiceSoftwareOptions]


class ColdStorageOptions(TypedDict, total=False):
    """Specifies the configuration for cold storage options such as enabled"""

    Enabled: Boolean


class ZoneAwarenessConfig(TypedDict, total=False):
    """The zone awareness configuration for the domain cluster, such as the
    number of availability zones.
    """

    AvailabilityZoneCount: Optional[IntegerClass]


class ClusterConfig(TypedDict, total=False):
    """The configuration for the domain cluster, such as the type and number of
    instances.
    """

    InstanceType: Optional[OpenSearchPartitionInstanceType]
    InstanceCount: Optional[IntegerClass]
    DedicatedMasterEnabled: Optional[Boolean]
    ZoneAwarenessEnabled: Optional[Boolean]
    ZoneAwarenessConfig: Optional[ZoneAwarenessConfig]
    DedicatedMasterType: Optional[OpenSearchPartitionInstanceType]
    DedicatedMasterCount: Optional[IntegerClass]
    WarmEnabled: Optional[Boolean]
    WarmType: Optional[OpenSearchWarmPartitionInstanceType]
    WarmCount: Optional[IntegerClass]
    ColdStorageOptions: Optional[ColdStorageOptions]


class ClusterConfigStatus(TypedDict, total=False):
    """The configuration status for the specified domain."""

    Options: ClusterConfig
    Status: OptionStatus


class CognitoOptions(TypedDict, total=False):
    """Options to specify the Cognito user and identity pools for OpenSearch
    Dashboards authentication. For more information, see `Configuring Amazon
    Cognito authentication for OpenSearch
    Dashboards <http://docs.aws.amazon.com/opensearch-service/latest/developerguide/cognito-auth.html>`__.
    """

    Enabled: Optional[Boolean]
    UserPoolId: Optional[UserPoolId]
    IdentityPoolId: Optional[IdentityPoolId]
    RoleArn: Optional[RoleArn]


class CognitoOptionsStatus(TypedDict, total=False):
    """The status of the Cognito options for the specified domain."""

    Options: CognitoOptions
    Status: OptionStatus


VersionList = List[VersionString]


class CompatibleVersionsMap(TypedDict, total=False):
    """A map from an ``EngineVersion`` to a list of compatible
    ``EngineVersion`` s to which the domain can be upgraded.
    """

    SourceVersion: Optional[VersionString]
    TargetVersions: Optional[VersionList]


CompatibleVersionsList = List[CompatibleVersionsMap]


class DomainEndpointOptions(TypedDict, total=False):
    """Options to configure the endpoint for the domain."""

    EnforceHTTPS: Optional[Boolean]
    TLSSecurityPolicy: Optional[TLSSecurityPolicy]
    CustomEndpointEnabled: Optional[Boolean]
    CustomEndpoint: Optional[DomainNameFqdn]
    CustomEndpointCertificateArn: Optional[ARN]


class LogPublishingOption(TypedDict, total=False):
    """| Log Publishing option that is set for a given domain.
    | Attributes and their details:

    -  CloudWatchLogsLogGroupArn: ARN of the Cloudwatch log group to publish
       logs to.
    -  Enabled: Whether the log publishing for a given log type is enabled
       or not.
    """

    CloudWatchLogsLogGroupArn: Optional[CloudWatchLogsLogGroupArn]
    Enabled: Optional[Boolean]


LogPublishingOptions = Dict[LogType, LogPublishingOption]


class NodeToNodeEncryptionOptions(TypedDict, total=False):
    """The node-to-node encryption options."""

    Enabled: Optional[Boolean]


class EncryptionAtRestOptions(TypedDict, total=False):
    """Specifies encryption at rest options."""

    Enabled: Optional[Boolean]
    KmsKeyId: Optional[KmsKeyId]


StringList = List[String]


class VPCOptions(TypedDict, total=False):
    """Options to specify the subnets and security groups for the VPC endpoint.
    For more information, see `Launching your Amazon OpenSearch Service
    domains using a
    VPC <http://docs.aws.amazon.com/opensearch-service/latest/developerguide/vpc.html>`__.
    """

    SubnetIds: Optional[StringList]
    SecurityGroupIds: Optional[StringList]


class SnapshotOptions(TypedDict, total=False):
    """The time, in UTC format, when the service takes a daily automated
    snapshot of the specified domain. Default is ``0`` hours.
    """

    AutomatedSnapshotStartHour: Optional[IntegerClass]


class EBSOptions(TypedDict, total=False):
    """Options to enable, disable, and specify the properties of EBS storage
    volumes.
    """

    EBSEnabled: Optional[Boolean]
    VolumeType: Optional[VolumeType]
    VolumeSize: Optional[IntegerClass]
    Iops: Optional[IntegerClass]


class CreateDomainRequest(ServiceRequest):
    DomainName: DomainName
    EngineVersion: Optional[VersionString]
    ClusterConfig: Optional[ClusterConfig]
    EBSOptions: Optional[EBSOptions]
    AccessPolicies: Optional[PolicyDocument]
    SnapshotOptions: Optional[SnapshotOptions]
    VPCOptions: Optional[VPCOptions]
    CognitoOptions: Optional[CognitoOptions]
    EncryptionAtRestOptions: Optional[EncryptionAtRestOptions]
    NodeToNodeEncryptionOptions: Optional[NodeToNodeEncryptionOptions]
    AdvancedOptions: Optional[AdvancedOptions]
    LogPublishingOptions: Optional[LogPublishingOptions]
    DomainEndpointOptions: Optional[DomainEndpointOptions]
    AdvancedSecurityOptions: Optional[AdvancedSecurityOptionsInput]
    TagList: Optional[TagList]
    AutoTuneOptions: Optional[AutoTuneOptionsInput]


class VPCDerivedInfo(TypedDict, total=False):
    """Options to specify the subnets and security groups for the VPC endpoint.
    For more information, see `Launching your Amazon OpenSearch Service
    domains using a
    VPC <http://docs.aws.amazon.com/opensearch-service/latest/developerguide/vpc.html>`__.
    """

    VPCId: Optional[String]
    SubnetIds: Optional[StringList]
    AvailabilityZones: Optional[StringList]
    SecurityGroupIds: Optional[StringList]


EndpointsMap = Dict[String, ServiceUrl]


class DomainStatus(TypedDict, total=False):
    """The current status of a domain."""

    DomainId: DomainId
    DomainName: DomainName
    ARN: ARN
    Created: Optional[Boolean]
    Deleted: Optional[Boolean]
    Endpoint: Optional[ServiceUrl]
    Endpoints: Optional[EndpointsMap]
    Processing: Optional[Boolean]
    UpgradeProcessing: Optional[Boolean]
    EngineVersion: Optional[VersionString]
    ClusterConfig: ClusterConfig
    EBSOptions: Optional[EBSOptions]
    AccessPolicies: Optional[PolicyDocument]
    SnapshotOptions: Optional[SnapshotOptions]
    VPCOptions: Optional[VPCDerivedInfo]
    CognitoOptions: Optional[CognitoOptions]
    EncryptionAtRestOptions: Optional[EncryptionAtRestOptions]
    NodeToNodeEncryptionOptions: Optional[NodeToNodeEncryptionOptions]
    AdvancedOptions: Optional[AdvancedOptions]
    LogPublishingOptions: Optional[LogPublishingOptions]
    ServiceSoftwareOptions: Optional[ServiceSoftwareOptions]
    DomainEndpointOptions: Optional[DomainEndpointOptions]
    AdvancedSecurityOptions: Optional[AdvancedSecurityOptions]
    AutoTuneOptions: Optional[AutoTuneOptionsOutput]


class CreateDomainResponse(TypedDict, total=False):
    """The result of a ``CreateDomain`` operation. Contains the status of the
    newly created Amazon OpenSearch Service domain.
    """

    DomainStatus: Optional[DomainStatus]


class CreateOutboundConnectionRequest(ServiceRequest):
    """Container for the parameters to the ``CreateOutboundConnection``
    operation.
    """

    LocalDomainInfo: DomainInformationContainer
    RemoteDomainInfo: DomainInformationContainer
    ConnectionAlias: ConnectionAlias


class OutboundConnectionStatus(TypedDict, total=False):
    """The connection status of an outbound cross-cluster connection."""

    StatusCode: Optional[OutboundConnectionStatusCode]
    Message: Optional[ConnectionStatusMessage]


class CreateOutboundConnectionResponse(TypedDict, total=False):
    """The result of a ``CreateOutboundConnection`` request. Contains the
    details about the newly created cross-cluster connection.
    """

    LocalDomainInfo: Optional[DomainInformationContainer]
    RemoteDomainInfo: Optional[DomainInformationContainer]
    ConnectionAlias: Optional[ConnectionAlias]
    ConnectionStatus: Optional[OutboundConnectionStatus]
    ConnectionId: Optional[ConnectionId]


class PackageSource(TypedDict, total=False):
    """The Amazon S3 location for importing the package specified as
    ``S3BucketName`` and ``S3Key``
    """

    S3BucketName: Optional[S3BucketName]
    S3Key: Optional[S3Key]


class CreatePackageRequest(ServiceRequest):
    """Container for request parameters to the ``CreatePackage`` operation."""

    PackageName: PackageName
    PackageType: PackageType
    PackageDescription: Optional[PackageDescription]
    PackageSource: PackageSource


CreatedAt = datetime


class PackageDetails(TypedDict, total=False):
    """Basic information about a package."""

    PackageID: Optional[PackageID]
    PackageName: Optional[PackageName]
    PackageType: Optional[PackageType]
    PackageDescription: Optional[PackageDescription]
    PackageStatus: Optional[PackageStatus]
    CreatedAt: Optional[CreatedAt]
    LastUpdatedAt: Optional[LastUpdated]
    AvailablePackageVersion: Optional[PackageVersion]
    ErrorDetails: Optional[ErrorDetails]


class CreatePackageResponse(TypedDict, total=False):
    """Container for the response returned by the ``CreatePackage`` operation."""

    PackageDetails: Optional[PackageDetails]


class DeleteDomainRequest(ServiceRequest):
    """Container for the parameters to the ``DeleteDomain`` operation.
    Specifies the name of the domain you want to delete.
    """

    DomainName: DomainName


class DeleteDomainResponse(TypedDict, total=False):
    """The result of a ``DeleteDomain`` request. Contains the status of the
    pending deletion, or a "domain not found" error if the domain and all of
    its resources have been deleted.
    """

    DomainStatus: Optional[DomainStatus]


class DeleteInboundConnectionRequest(ServiceRequest):
    """Container for the parameters to the ``DeleteInboundConnection``
    operation.
    """

    ConnectionId: ConnectionId


class DeleteInboundConnectionResponse(TypedDict, total=False):
    """The result of a ``DeleteInboundConnection`` operation. Contains details
    about the deleted inbound connection.
    """

    Connection: Optional[InboundConnection]


class DeleteOutboundConnectionRequest(ServiceRequest):
    """Container for the parameters to the ``DeleteOutboundConnection``
    operation.
    """

    ConnectionId: ConnectionId


class OutboundConnection(TypedDict, total=False):
    """Specifies details about an outbound connection."""

    LocalDomainInfo: Optional[DomainInformationContainer]
    RemoteDomainInfo: Optional[DomainInformationContainer]
    ConnectionId: Optional[ConnectionId]
    ConnectionAlias: Optional[ConnectionAlias]
    ConnectionStatus: Optional[OutboundConnectionStatus]


class DeleteOutboundConnectionResponse(TypedDict, total=False):
    """The result of a ``DeleteOutboundConnection`` operation. Contains details
    about the deleted outbound connection.
    """

    Connection: Optional[OutboundConnection]


class DeletePackageRequest(ServiceRequest):
    """Container for the request parameters to the ``DeletePackage`` operation."""

    PackageID: PackageID


class DeletePackageResponse(TypedDict, total=False):
    """Container for the response parameters to the ``DeletePackage``
    operation.
    """

    PackageDetails: Optional[PackageDetails]


class DescribeDomainAutoTunesRequest(ServiceRequest):
    """Container for the parameters to the ``DescribeDomainAutoTunes``
    operation.
    """

    DomainName: DomainName
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class DescribeDomainAutoTunesResponse(TypedDict, total=False):
    """The result of a ``DescribeDomainAutoTunes`` request. See `Auto-Tune for
    Amazon OpenSearch
    Service <https://docs.aws.amazon.com/opensearch-service/latest/developerguide/auto-tune.html>`__
    for more information.
    """

    AutoTunes: Optional[AutoTuneList]
    NextToken: Optional[NextToken]


class DescribeDomainConfigRequest(ServiceRequest):
    """Container for the parameters to the ``DescribeDomainConfig`` operation.
    Specifies the domain name for which you want configuration information.
    """

    DomainName: DomainName


class DomainEndpointOptionsStatus(TypedDict, total=False):
    """The configured endpoint options for the domain and their current status."""

    Options: DomainEndpointOptions
    Status: OptionStatus


class LogPublishingOptionsStatus(TypedDict, total=False):
    """The configured log publishing options for the domain and their current
    status.
    """

    Options: Optional[LogPublishingOptions]
    Status: Optional[OptionStatus]


class NodeToNodeEncryptionOptionsStatus(TypedDict, total=False):
    """Status of the node-to-node encryption options for the specified domain."""

    Options: NodeToNodeEncryptionOptions
    Status: OptionStatus


class EncryptionAtRestOptionsStatus(TypedDict, total=False):
    """Status of the encryption At Rest options for the specified domain."""

    Options: EncryptionAtRestOptions
    Status: OptionStatus


class VPCDerivedInfoStatus(TypedDict, total=False):
    """Status of the VPC options for the specified domain."""

    Options: VPCDerivedInfo
    Status: OptionStatus


class SnapshotOptionsStatus(TypedDict, total=False):
    """Status of a daily automated snapshot."""

    Options: SnapshotOptions
    Status: OptionStatus


class EBSOptionsStatus(TypedDict, total=False):
    """Status of the EBS options for the specified domain."""

    Options: EBSOptions
    Status: OptionStatus


class VersionStatus(TypedDict, total=False):
    """The status of the OpenSearch version options for the specified
    OpenSearch domain.
    """

    Options: VersionString
    Status: OptionStatus


class DomainConfig(TypedDict, total=False):
    """The configuration of a domain."""

    EngineVersion: Optional[VersionStatus]
    ClusterConfig: Optional[ClusterConfigStatus]
    EBSOptions: Optional[EBSOptionsStatus]
    AccessPolicies: Optional[AccessPoliciesStatus]
    SnapshotOptions: Optional[SnapshotOptionsStatus]
    VPCOptions: Optional[VPCDerivedInfoStatus]
    CognitoOptions: Optional[CognitoOptionsStatus]
    EncryptionAtRestOptions: Optional[EncryptionAtRestOptionsStatus]
    NodeToNodeEncryptionOptions: Optional[NodeToNodeEncryptionOptionsStatus]
    AdvancedOptions: Optional[AdvancedOptionsStatus]
    LogPublishingOptions: Optional[LogPublishingOptionsStatus]
    DomainEndpointOptions: Optional[DomainEndpointOptionsStatus]
    AdvancedSecurityOptions: Optional[AdvancedSecurityOptionsStatus]
    AutoTuneOptions: Optional[AutoTuneOptionsStatus]


class DescribeDomainConfigResponse(TypedDict, total=False):
    """The result of a ``DescribeDomainConfig`` request. Contains the
    configuration information of the requested domain.
    """

    DomainConfig: DomainConfig


class DescribeDomainRequest(ServiceRequest):
    """Container for the parameters to the ``DescribeDomain`` operation."""

    DomainName: DomainName


class DescribeDomainResponse(TypedDict, total=False):
    """The result of a ``DescribeDomain`` request. Contains the status of the
    domain specified in the request.
    """

    DomainStatus: DomainStatus


DomainNameList = List[DomainName]


class DescribeDomainsRequest(ServiceRequest):
    """Container for the parameters to the ``DescribeDomains`` operation. By
    default, the API returns the status of all domains.
    """

    DomainNames: DomainNameList


DomainStatusList = List[DomainStatus]


class DescribeDomainsResponse(TypedDict, total=False):
    """The result of a ``DescribeDomains`` request. Contains the status of the
    specified domains or all domains owned by the account.
    """

    DomainStatusList: DomainStatusList


ValueStringList = List[NonEmptyString]


class Filter(TypedDict, total=False):
    """A filter used to limit results when describing inbound or outbound
    cross-cluster connections. Multiple values can be specified per filter.
    A cross-cluster connection must match at least one of the specified
    values for it to be returned from an operation.
    """

    Name: Optional[NonEmptyString]
    Values: Optional[ValueStringList]


FilterList = List[Filter]


class DescribeInboundConnectionsRequest(ServiceRequest):
    """Container for the parameters to the ``DescribeInboundConnections``
    operation.
    """

    Filters: Optional[FilterList]
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


InboundConnections = List[InboundConnection]


class DescribeInboundConnectionsResponse(TypedDict, total=False):
    """The result of a ``DescribeInboundConnections`` request. Contains a list
    of connections matching the filter criteria.
    """

    Connections: Optional[InboundConnections]
    NextToken: Optional[NextToken]


class DescribeInstanceTypeLimitsRequest(ServiceRequest):
    """Container for the parameters to the ``DescribeInstanceTypeLimits``
    operation.
    """

    DomainName: Optional[DomainName]
    InstanceType: OpenSearchPartitionInstanceType
    EngineVersion: VersionString


class InstanceCountLimits(TypedDict, total=False):
    """InstanceCountLimits represents the limits on the number of instances
    that can be created in Amazon OpenSearch Service for a given
    InstanceType.
    """

    MinimumInstanceCount: Optional[MinimumInstanceCount]
    MaximumInstanceCount: Optional[MaximumInstanceCount]


class InstanceLimits(TypedDict, total=False):
    """InstanceLimits represents the list of instance-related attributes that
    are available for a given InstanceType.
    """

    InstanceCountLimits: Optional[InstanceCountLimits]


class StorageTypeLimit(TypedDict, total=False):
    """Limits that are applicable for the given storage type."""

    LimitName: Optional[LimitName]
    LimitValues: Optional[LimitValueList]


StorageTypeLimitList = List[StorageTypeLimit]


class StorageType(TypedDict, total=False):
    """StorageTypes represents the list of storage-related types and their
    attributes that are available for a given InstanceType.
    """

    StorageTypeName: Optional[StorageTypeName]
    StorageSubTypeName: Optional[StorageSubTypeName]
    StorageTypeLimits: Optional[StorageTypeLimitList]


StorageTypeList = List[StorageType]


class Limits(TypedDict, total=False):
    """| Limits for a given InstanceType and for each of its roles.
    | Limits contains the following: ``StorageTypes``, ``InstanceLimits``,
      and ``AdditionalLimits``
    """

    StorageTypes: Optional[StorageTypeList]
    InstanceLimits: Optional[InstanceLimits]
    AdditionalLimits: Optional[AdditionalLimitList]


LimitsByRole = Dict[InstanceRole, Limits]


class DescribeInstanceTypeLimitsResponse(TypedDict, total=False):
    """Container for the parameters received from the
    ``DescribeInstanceTypeLimits`` operation.
    """

    LimitsByRole: Optional[LimitsByRole]


class DescribeOutboundConnectionsRequest(ServiceRequest):
    """Container for the parameters to the ``DescribeOutboundConnections``
    operation.
    """

    Filters: Optional[FilterList]
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


OutboundConnections = List[OutboundConnection]


class DescribeOutboundConnectionsResponse(TypedDict, total=False):
    """The result of a ``DescribeOutboundConnections`` request. Contains the
    list of connections matching the filter criteria.
    """

    Connections: Optional[OutboundConnections]
    NextToken: Optional[NextToken]


DescribePackagesFilterValues = List[DescribePackagesFilterValue]


class DescribePackagesFilter(TypedDict, total=False):
    """A filter to apply to the ``DescribePackage`` response."""

    Name: Optional[DescribePackagesFilterName]
    Value: Optional[DescribePackagesFilterValues]


DescribePackagesFilterList = List[DescribePackagesFilter]


class DescribePackagesRequest(ServiceRequest):
    """Container for the request parameters to the ``DescribePackage``
    operation.
    """

    Filters: Optional[DescribePackagesFilterList]
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


PackageDetailsList = List[PackageDetails]


class DescribePackagesResponse(TypedDict, total=False):
    """Container for the response returned by the ``DescribePackages``
    operation.
    """

    PackageDetailsList: Optional[PackageDetailsList]
    NextToken: Optional[String]


class DescribeReservedInstanceOfferingsRequest(ServiceRequest):
    """Container for parameters to ``DescribeReservedInstanceOfferings``"""

    ReservedInstanceOfferingId: Optional[GUID]
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class RecurringCharge(TypedDict, total=False):
    """Contains the specific price and frequency of a recurring charges for a
    reserved OpenSearch instance, or for a reserved OpenSearch instance
    offering.
    """

    RecurringChargeAmount: Optional[Double]
    RecurringChargeFrequency: Optional[String]


RecurringChargeList = List[RecurringCharge]


class ReservedInstanceOffering(TypedDict, total=False):
    """Details of a reserved OpenSearch instance offering."""

    ReservedInstanceOfferingId: Optional[GUID]
    InstanceType: Optional[OpenSearchPartitionInstanceType]
    Duration: Optional[Integer]
    FixedPrice: Optional[Double]
    UsagePrice: Optional[Double]
    CurrencyCode: Optional[String]
    PaymentOption: Optional[ReservedInstancePaymentOption]
    RecurringCharges: Optional[RecurringChargeList]


ReservedInstanceOfferingList = List[ReservedInstanceOffering]


class DescribeReservedInstanceOfferingsResponse(TypedDict, total=False):
    """Container for results from ``DescribeReservedInstanceOfferings``"""

    NextToken: Optional[NextToken]
    ReservedInstanceOfferings: Optional[ReservedInstanceOfferingList]


class DescribeReservedInstancesRequest(ServiceRequest):
    """Container for parameters to ``DescribeReservedInstances``"""

    ReservedInstanceId: Optional[GUID]
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


Long = int


class ReservedInstance(TypedDict, total=False):
    """Details of a reserved OpenSearch instance."""

    ReservationName: Optional[ReservationToken]
    ReservedInstanceId: Optional[GUID]
    BillingSubscriptionId: Optional[Long]
    ReservedInstanceOfferingId: Optional[String]
    InstanceType: Optional[OpenSearchPartitionInstanceType]
    StartTime: Optional[UpdateTimestamp]
    Duration: Optional[Integer]
    FixedPrice: Optional[Double]
    UsagePrice: Optional[Double]
    CurrencyCode: Optional[String]
    InstanceCount: Optional[Integer]
    State: Optional[String]
    PaymentOption: Optional[ReservedInstancePaymentOption]
    RecurringCharges: Optional[RecurringChargeList]


ReservedInstanceList = List[ReservedInstance]


class DescribeReservedInstancesResponse(TypedDict, total=False):
    """Container for results from ``DescribeReservedInstances``"""

    NextToken: Optional[String]
    ReservedInstances: Optional[ReservedInstanceList]


class DissociatePackageRequest(ServiceRequest):
    """Container for the request parameters to the ``DissociatePackage``
    operation.
    """

    PackageID: PackageID
    DomainName: DomainName


class DissociatePackageResponse(TypedDict, total=False):
    """Container for the response returned by ``DissociatePackage`` operation."""

    DomainPackageDetails: Optional[DomainPackageDetails]


class DomainInfo(TypedDict, total=False):
    DomainName: Optional[DomainName]
    EngineType: Optional[EngineType]


DomainInfoList = List[DomainInfo]
DomainPackageDetailsList = List[DomainPackageDetails]


class DryRunResults(TypedDict, total=False):
    DeploymentType: Optional[DeploymentType]
    Message: Optional[Message]


class GetCompatibleVersionsRequest(ServiceRequest):
    """Container for the request parameters to ``GetCompatibleVersions``
    operation.
    """

    DomainName: Optional[DomainName]


class GetCompatibleVersionsResponse(TypedDict, total=False):
    """Container for the response returned by the ``GetCompatibleVersions``
    operation.
    """

    CompatibleVersions: Optional[CompatibleVersionsList]


class GetPackageVersionHistoryRequest(ServiceRequest):
    """Container for the request parameters to the ``GetPackageVersionHistory``
    operation.
    """

    PackageID: PackageID
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class PackageVersionHistory(TypedDict, total=False):
    """Details of a package version."""

    PackageVersion: Optional[PackageVersion]
    CommitMessage: Optional[CommitMessage]
    CreatedAt: Optional[CreatedAt]


PackageVersionHistoryList = List[PackageVersionHistory]


class GetPackageVersionHistoryResponse(TypedDict, total=False):
    """Container for response returned by ``GetPackageVersionHistory``
    operation.
    """

    PackageID: Optional[PackageID]
    PackageVersionHistoryList: Optional[PackageVersionHistoryList]
    NextToken: Optional[String]


class GetUpgradeHistoryRequest(ServiceRequest):
    """Container for the request parameters to the ``GetUpgradeHistory``
    operation.
    """

    DomainName: DomainName
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


Issues = List[Issue]


class UpgradeStepItem(TypedDict, total=False):
    """Represents a single step of the upgrade or upgrade eligibility check
    workflow.
    """

    UpgradeStep: Optional[UpgradeStep]
    UpgradeStepStatus: Optional[UpgradeStatus]
    Issues: Optional[Issues]
    ProgressPercent: Optional[Double]


UpgradeStepsList = List[UpgradeStepItem]
StartTimestamp = datetime


class UpgradeHistory(TypedDict, total=False):
    """History of the last 10 upgrades and upgrade eligibility checks."""

    UpgradeName: Optional[UpgradeName]
    StartTimestamp: Optional[StartTimestamp]
    UpgradeStatus: Optional[UpgradeStatus]
    StepsList: Optional[UpgradeStepsList]


UpgradeHistoryList = List[UpgradeHistory]


class GetUpgradeHistoryResponse(TypedDict, total=False):
    """Container for the response returned by the ``GetUpgradeHistory``
    operation.
    """

    UpgradeHistories: Optional[UpgradeHistoryList]
    NextToken: Optional[String]


class GetUpgradeStatusRequest(ServiceRequest):
    """Container for the request parameters to the ``GetUpgradeStatus``
    operation.
    """

    DomainName: DomainName


class GetUpgradeStatusResponse(TypedDict, total=False):
    """Container for the response returned by the ``GetUpgradeStatus``
    operation.
    """

    UpgradeStep: Optional[UpgradeStep]
    StepStatus: Optional[UpgradeStatus]
    UpgradeName: Optional[UpgradeName]


InstanceRoleList = List[InstanceRole]


class InstanceTypeDetails(TypedDict, total=False):
    InstanceType: Optional[OpenSearchPartitionInstanceType]
    EncryptionEnabled: Optional[Boolean]
    CognitoEnabled: Optional[Boolean]
    AppLogsEnabled: Optional[Boolean]
    AdvancedSecurityEnabled: Optional[Boolean]
    WarmEnabled: Optional[Boolean]
    InstanceRole: Optional[InstanceRoleList]


InstanceTypeDetailsList = List[InstanceTypeDetails]


class ListDomainNamesRequest(ServiceRequest):
    """Container for the parameters to the ``ListDomainNames`` operation."""

    EngineType: Optional[EngineType]


class ListDomainNamesResponse(TypedDict, total=False):
    """The result of a ``ListDomainNames`` operation. Contains the names of all
    domains owned by this account and their respective engine types.
    """

    DomainNames: Optional[DomainInfoList]


class ListDomainsForPackageRequest(ServiceRequest):
    """Container for the request parameters to the ``ListDomainsForPackage``
    operation.
    """

    PackageID: PackageID
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class ListDomainsForPackageResponse(TypedDict, total=False):
    """Container for the response parameters to the ``ListDomainsForPackage``
    operation.
    """

    DomainPackageDetailsList: Optional[DomainPackageDetailsList]
    NextToken: Optional[String]


class ListInstanceTypeDetailsRequest(ServiceRequest):
    EngineVersion: VersionString
    DomainName: Optional[DomainName]
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class ListInstanceTypeDetailsResponse(TypedDict, total=False):
    InstanceTypeDetails: Optional[InstanceTypeDetailsList]
    NextToken: Optional[NextToken]


class ListPackagesForDomainRequest(ServiceRequest):
    """Container for the request parameters to the ``ListPackagesForDomain``
    operation.
    """

    DomainName: DomainName
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class ListPackagesForDomainResponse(TypedDict, total=False):
    """Container for the response parameters to the ``ListPackagesForDomain``
    operation.
    """

    DomainPackageDetailsList: Optional[DomainPackageDetailsList]
    NextToken: Optional[String]


class ListTagsRequest(ServiceRequest):
    """Container for the parameters to the ``ListTags`` operation. Specify the
    ``ARN`` of the domain that the tags you want to view are attached to.
    """

    ARN: ARN


class ListTagsResponse(TypedDict, total=False):
    """The result of a ``ListTags`` operation. Contains tags for all requested
    domains.
    """

    TagList: Optional[TagList]


class ListVersionsRequest(ServiceRequest):
    """Container for the parameters to the ``ListVersions`` operation.

    Use ``MaxResults`` to control the maximum number of results to retrieve
    in a single call.

    Use ``NextToken`` in response to retrieve more results. If the received
    response does not contain a NextToken, there are no more results to
    retrieve.
    """

    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class ListVersionsResponse(TypedDict, total=False):
    """Container for the parameters for response received from the
    ``ListVersions`` operation.
    """

    Versions: Optional[VersionList]
    NextToken: Optional[NextToken]


class PurchaseReservedInstanceOfferingRequest(ServiceRequest):
    """Container for parameters to ``PurchaseReservedInstanceOffering``"""

    ReservedInstanceOfferingId: GUID
    ReservationName: ReservationToken
    InstanceCount: Optional[InstanceCount]


class PurchaseReservedInstanceOfferingResponse(TypedDict, total=False):
    """Represents the output of a ``PurchaseReservedInstanceOffering``
    operation.
    """

    ReservedInstanceId: Optional[GUID]
    ReservationName: Optional[ReservationToken]


class RejectInboundConnectionRequest(ServiceRequest):
    """Container for the parameters to the ``RejectInboundConnection``
    operation.
    """

    ConnectionId: ConnectionId


class RejectInboundConnectionResponse(TypedDict, total=False):
    """The result of a ``RejectInboundConnection`` operation. Contains details
    about the rejected inbound connection.
    """

    Connection: Optional[InboundConnection]


class RemoveTagsRequest(ServiceRequest):
    """Container for the parameters to the ``RemoveTags`` operation. Specify
    the ``ARN`` for the domain from which you want to remove the specified
    ``TagKey``.
    """

    ARN: ARN
    TagKeys: StringList


class StartServiceSoftwareUpdateRequest(ServiceRequest):
    """Container for the parameters to the ``StartServiceSoftwareUpdate``
    operation. Specifies the name of the domain to schedule a service
    software update for.
    """

    DomainName: DomainName


class StartServiceSoftwareUpdateResponse(TypedDict, total=False):
    """The result of a ``StartServiceSoftwareUpdate`` operation. Contains the
    status of the update.
    """

    ServiceSoftwareOptions: Optional[ServiceSoftwareOptions]


class UpdateDomainConfigRequest(ServiceRequest):
    """Container for the parameters to the ``UpdateDomain`` operation.
    Specifies the type and number of instances in the domain cluster.
    """

    DomainName: DomainName
    ClusterConfig: Optional[ClusterConfig]
    EBSOptions: Optional[EBSOptions]
    SnapshotOptions: Optional[SnapshotOptions]
    VPCOptions: Optional[VPCOptions]
    CognitoOptions: Optional[CognitoOptions]
    AdvancedOptions: Optional[AdvancedOptions]
    AccessPolicies: Optional[PolicyDocument]
    LogPublishingOptions: Optional[LogPublishingOptions]
    EncryptionAtRestOptions: Optional[EncryptionAtRestOptions]
    DomainEndpointOptions: Optional[DomainEndpointOptions]
    NodeToNodeEncryptionOptions: Optional[NodeToNodeEncryptionOptions]
    AdvancedSecurityOptions: Optional[AdvancedSecurityOptionsInput]
    AutoTuneOptions: Optional[AutoTuneOptions]
    DryRun: Optional[DryRun]


class UpdateDomainConfigResponse(TypedDict, total=False):
    """The result of an ``UpdateDomain`` request. Contains the status of the
    domain being updated.
    """

    DomainConfig: DomainConfig
    DryRunResults: Optional[DryRunResults]


class UpdatePackageRequest(ServiceRequest):
    """Container for request parameters to the ``UpdatePackage`` operation."""

    PackageID: PackageID
    PackageSource: PackageSource
    PackageDescription: Optional[PackageDescription]
    CommitMessage: Optional[CommitMessage]


class UpdatePackageResponse(TypedDict, total=False):
    """Container for the response returned by the ``UpdatePackage`` operation."""

    PackageDetails: Optional[PackageDetails]


class UpgradeDomainRequest(ServiceRequest):
    """Container for the request parameters to ``UpgradeDomain`` operation."""

    DomainName: DomainName
    TargetVersion: VersionString
    PerformCheckOnly: Optional[Boolean]
    AdvancedOptions: Optional[AdvancedOptions]


class UpgradeDomainResponse(TypedDict, total=False):
    """Container for response returned by ``UpgradeDomain`` operation."""

    UpgradeId: Optional[String]
    DomainName: Optional[DomainName]
    TargetVersion: Optional[VersionString]
    PerformCheckOnly: Optional[Boolean]
    AdvancedOptions: Optional[AdvancedOptions]


class OpensearchApi:

    service = "opensearch"
    version = "2021-01-01"

    @handler("AcceptInboundConnection")
    def accept_inbound_connection(
        self, context: RequestContext, connection_id: ConnectionId
    ) -> AcceptInboundConnectionResponse:
        """Allows the remote domain owner to accept an inbound cross-cluster
        connection request.

        :param connection_id: The ID of the inbound connection you want to accept.
        :returns: AcceptInboundConnectionResponse
        :raises ResourceNotFoundException:
        :raises LimitExceededException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("AddTags")
    def add_tags(self, context: RequestContext, arn: ARN, tag_list: TagList) -> None:
        """Attaches tags to an existing domain. Tags are a set of case-sensitive
        key value pairs. An domain can have up to 10 tags. See `Tagging Amazon
        OpenSearch Service
        domains <http://docs.aws.amazon.com/opensearch-service/latest/developerguide/managedomains.html#managedomains-awsresorcetagging>`__
        for more information.

        :param arn: Specify the ``ARN`` of the domain you want to add tags to.
        :param tag_list: List of ``Tag`` to add to the domain.
        :raises BaseException:
        :raises LimitExceededException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("AssociatePackage")
    def associate_package(
        self, context: RequestContext, package_id: PackageID, domain_name: DomainName
    ) -> AssociatePackageResponse:
        """Associates a package with an Amazon OpenSearch Service domain.

        :param package_id: Internal ID of the package to associate with a domain.
        :param domain_name: The name of the domain to associate the package with.
        :returns: AssociatePackageResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        :raises ConflictException:
        """
        raise NotImplementedError

    @handler("CancelServiceSoftwareUpdate")
    def cancel_service_software_update(
        self, context: RequestContext, domain_name: DomainName
    ) -> CancelServiceSoftwareUpdateResponse:
        """Cancels a scheduled service software update for an Amazon OpenSearch
        Service domain. You can only perform this operation before the
        ``AutomatedUpdateDate`` and when the ``UpdateStatus`` is in the
        ``PENDING_UPDATE`` state.

        :param domain_name: The name of the domain that you want to stop the latest service software
        update on.
        :returns: CancelServiceSoftwareUpdateResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("CreateDomain")
    def create_domain(
        self,
        context: RequestContext,
        domain_name: DomainName,
        engine_version: VersionString = None,
        cluster_config: ClusterConfig = None,
        ebs_options: EBSOptions = None,
        access_policies: PolicyDocument = None,
        snapshot_options: SnapshotOptions = None,
        vpc_options: VPCOptions = None,
        cognito_options: CognitoOptions = None,
        encryption_at_rest_options: EncryptionAtRestOptions = None,
        node_to_node_encryption_options: NodeToNodeEncryptionOptions = None,
        advanced_options: AdvancedOptions = None,
        log_publishing_options: LogPublishingOptions = None,
        domain_endpoint_options: DomainEndpointOptions = None,
        advanced_security_options: AdvancedSecurityOptionsInput = None,
        tag_list: TagList = None,
        auto_tune_options: AutoTuneOptionsInput = None,
    ) -> CreateDomainResponse:
        """Creates a new Amazon OpenSearch Service domain. For more information,
        see `Creating and managing Amazon OpenSearch Service
        domains <http://docs.aws.amazon.com/opensearch-service/latest/developerguide/createupdatedomains.html>`__
        in the *Amazon OpenSearch Service Developer Guide*.

        :param domain_name: The name of the Amazon OpenSearch Service domain you're creating.
        :param engine_version: String of format Elasticsearch_X.
        :param cluster_config: Configuration options for a domain.
        :param ebs_options: Options to enable, disable, and specify the type and size of EBS storage
        volumes.
        :param access_policies: IAM access policy as a JSON-formatted string.
        :param snapshot_options: Option to set time, in UTC format, of the daily automated snapshot.
        :param vpc_options: Options to specify the subnets and security groups for a VPC endpoint.
        :param cognito_options: Options to specify the Cognito user and identity pools for OpenSearch
        Dashboards authentication.
        :param encryption_at_rest_options: Options for encryption of data at rest.
        :param node_to_node_encryption_options: Node-to-node encryption options.
        :param advanced_options: Option to allow references to indices in an HTTP request body.
        :param log_publishing_options: Map of ``LogType`` and ``LogPublishingOption``, each containing options
        to publish a given type of OpenSearch log.
        :param domain_endpoint_options: Options to specify configurations that will be applied to the domain
        endpoint.
        :param advanced_security_options: Specifies advanced security options.
        :param tag_list: A list of ``Tag`` added during domain creation.
        :param auto_tune_options: Specifies Auto-Tune options.
        :returns: CreateDomainResponse
        :raises BaseException:
        :raises DisabledOperationException:
        :raises InternalException:
        :raises InvalidTypeException:
        :raises LimitExceededException:
        :raises ResourceAlreadyExistsException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("CreateOutboundConnection")
    def create_outbound_connection(
        self,
        context: RequestContext,
        local_domain_info: DomainInformationContainer,
        remote_domain_info: DomainInformationContainer,
        connection_alias: ConnectionAlias,
    ) -> CreateOutboundConnectionResponse:
        """Creates a new cross-cluster connection from a local OpenSearch domain to
        a remote OpenSearch domain.

        :param local_domain_info: The ``AWSDomainInformation`` for the local OpenSearch domain.
        :param remote_domain_info: The ``AWSDomainInformation`` for the remote OpenSearch domain.
        :param connection_alias: The connection alias used used by the customer for this cross-cluster
        connection.
        :returns: CreateOutboundConnectionResponse
        :raises LimitExceededException:
        :raises InternalException:
        :raises ResourceAlreadyExistsException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("CreatePackage")
    def create_package(
        self,
        context: RequestContext,
        package_name: PackageName,
        package_type: PackageType,
        package_source: PackageSource,
        package_description: PackageDescription = None,
    ) -> CreatePackageResponse:
        """Create a package for use with Amazon OpenSearch Service domains.

        :param package_name: Unique identifier for the package.
        :param package_type: Type of package.
        :param package_source: The Amazon S3 location from which to import the package.
        :param package_description: Description of the package.
        :returns: CreatePackageResponse
        :raises BaseException:
        :raises InternalException:
        :raises LimitExceededException:
        :raises InvalidTypeException:
        :raises ResourceAlreadyExistsException:
        :raises AccessDeniedException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DeleteDomain")
    def delete_domain(
        self, context: RequestContext, domain_name: DomainName
    ) -> DeleteDomainResponse:
        """Permanently deletes the specified domain and all of its data. Once a
        domain is deleted, it cannot be recovered.

        :param domain_name: The name of the domain you want to permanently delete.
        :returns: DeleteDomainResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DeleteInboundConnection")
    def delete_inbound_connection(
        self, context: RequestContext, connection_id: ConnectionId
    ) -> DeleteInboundConnectionResponse:
        """Allows the remote domain owner to delete an existing inbound
        cross-cluster connection.

        :param connection_id: The ID of the inbound connection to permanently delete.
        :returns: DeleteInboundConnectionResponse
        :raises ResourceNotFoundException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("DeleteOutboundConnection")
    def delete_outbound_connection(
        self, context: RequestContext, connection_id: ConnectionId
    ) -> DeleteOutboundConnectionResponse:
        """Allows the local domain owner to delete an existing outbound
        cross-cluster connection.

        :param connection_id: The ID of the outbound connection you want to permanently delete.
        :returns: DeleteOutboundConnectionResponse
        :raises ResourceNotFoundException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("DeletePackage")
    def delete_package(
        self, context: RequestContext, package_id: PackageID
    ) -> DeletePackageResponse:
        """Deletes the package.

        :param package_id: The internal ID of the package you want to delete.
        :returns: DeletePackageResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        :raises ConflictException:
        """
        raise NotImplementedError

    @handler("DescribeDomain")
    def describe_domain(
        self, context: RequestContext, domain_name: DomainName
    ) -> DescribeDomainResponse:
        """Returns domain configuration information about the specified domain,
        including the domain ID, domain endpoint, and domain ARN.

        :param domain_name: The name of the domain for which you want information.
        :returns: DescribeDomainResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DescribeDomainAutoTunes")
    def describe_domain_auto_tunes(
        self,
        context: RequestContext,
        domain_name: DomainName,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> DescribeDomainAutoTunesResponse:
        """Provides scheduled Auto-Tune action details for the domain, such as
        Auto-Tune action type, description, severity, and scheduled date.

        :param domain_name: The domain name for which you want Auto-Tune action details.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: NextToken is sent in case the earlier API call results contain the
        NextToken.
        :returns: DescribeDomainAutoTunesResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DescribeDomainConfig")
    def describe_domain_config(
        self, context: RequestContext, domain_name: DomainName
    ) -> DescribeDomainConfigResponse:
        """Provides cluster configuration information about the specified domain,
        such as the state, creation date, update version, and update date for
        cluster options.

        :param domain_name: The domain you want to get information about.
        :returns: DescribeDomainConfigResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DescribeDomains")
    def describe_domains(
        self, context: RequestContext, domain_names: DomainNameList
    ) -> DescribeDomainsResponse:
        """Returns domain configuration information about the specified domains,
        including the domain ID, domain endpoint, and domain ARN.

        :param domain_names: The domains for which you want information.
        :returns: DescribeDomainsResponse
        :raises BaseException:
        :raises InternalException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DescribeInboundConnections")
    def describe_inbound_connections(
        self,
        context: RequestContext,
        filters: FilterList = None,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> DescribeInboundConnectionsResponse:
        """Lists all the inbound cross-cluster connections for a remote domain.

        :param filters: A list of filters used to match properties for inbound cross-cluster
        connections.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: If more results are available and NextToken is present, make the next
        request to the same API with the received NextToken to paginate the
        remaining results.
        :returns: DescribeInboundConnectionsResponse
        :raises InvalidPaginationTokenException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("DescribeInstanceTypeLimits")
    def describe_instance_type_limits(
        self,
        context: RequestContext,
        instance_type: OpenSearchPartitionInstanceType,
        engine_version: VersionString,
        domain_name: DomainName = None,
    ) -> DescribeInstanceTypeLimitsResponse:
        """Describe the limits for a given instance type and OpenSearch or
        Elasticsearch version. When modifying an existing domain, specify the
        ``DomainName`` to see which limits you can modify.

        :param instance_type: The instance type for an OpenSearch cluster for which OpenSearch
        ``Limits`` are needed.
        :param engine_version: Version of OpenSearch for which ``Limits`` are needed.
        :param domain_name: The name of the domain you want to modify.
        :returns: DescribeInstanceTypeLimitsResponse
        :raises BaseException:
        :raises InternalException:
        :raises InvalidTypeException:
        :raises LimitExceededException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DescribeOutboundConnections")
    def describe_outbound_connections(
        self,
        context: RequestContext,
        filters: FilterList = None,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> DescribeOutboundConnectionsResponse:
        """Lists all the outbound cross-cluster connections for a local domain.

        :param filters: A list of filters used to match properties for outbound cross-cluster
        connections.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: NextToken is sent in case the earlier API call results contain the
        NextToken parameter.
        :returns: DescribeOutboundConnectionsResponse
        :raises InvalidPaginationTokenException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("DescribePackages")
    def describe_packages(
        self,
        context: RequestContext,
        filters: DescribePackagesFilterList = None,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> DescribePackagesResponse:
        """Describes all packages available to Amazon OpenSearch Service domains.
        Includes options for filtering, limiting the number of results, and
        pagination.

        :param filters: Only returns packages that match the ``DescribePackagesFilterList``
        values.
        :param max_results: Limits results to a maximum number of packages.
        :param next_token: Used for pagination.
        :returns: DescribePackagesResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DescribeReservedInstanceOfferings")
    def describe_reserved_instance_offerings(
        self,
        context: RequestContext,
        reserved_instance_offering_id: GUID = None,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> DescribeReservedInstanceOfferingsResponse:
        """Lists available reserved OpenSearch instance offerings.

        :param reserved_instance_offering_id: The offering identifier filter value.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: Provides an identifier to allow retrieval of paginated results.
        :returns: DescribeReservedInstanceOfferingsResponse
        :raises ResourceNotFoundException:
        :raises ValidationException:
        :raises DisabledOperationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("DescribeReservedInstances")
    def describe_reserved_instances(
        self,
        context: RequestContext,
        reserved_instance_id: GUID = None,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> DescribeReservedInstancesResponse:
        """Returns information about reserved OpenSearch instances for this
        account.

        :param reserved_instance_id: The reserved instance identifier filter value.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: Provides an identifier to allow retrieval of paginated results.
        :returns: DescribeReservedInstancesResponse
        :raises ResourceNotFoundException:
        :raises InternalException:
        :raises ValidationException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("DissociatePackage")
    def dissociate_package(
        self, context: RequestContext, package_id: PackageID, domain_name: DomainName
    ) -> DissociatePackageResponse:
        """Dissociates a package from the Amazon OpenSearch Service domain.

        :param package_id: The internal ID of the package to associate with a domain.
        :param domain_name: The name of the domain to associate the package with.
        :returns: DissociatePackageResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        :raises ConflictException:
        """
        raise NotImplementedError

    @handler("GetCompatibleVersions")
    def get_compatible_versions(
        self, context: RequestContext, domain_name: DomainName = None
    ) -> GetCompatibleVersionsResponse:
        """Returns a list of upgrade-compatible versions of
        OpenSearch/Elasticsearch. You can optionally pass a ``DomainName`` to
        get all upgrade-compatible versions of OpenSearch/Elasticsearch for that
        specific domain.

        :param domain_name: The name of an domain.
        :returns: GetCompatibleVersionsResponse
        :raises BaseException:
        :raises ResourceNotFoundException:
        :raises DisabledOperationException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("GetPackageVersionHistory")
    def get_package_version_history(
        self,
        context: RequestContext,
        package_id: PackageID,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> GetPackageVersionHistoryResponse:
        """Returns a list of package versions, along with their creation time and
        commit message.

        :param package_id: Returns an audit history of package versions.
        :param max_results: Limits results to a maximum number of package versions.
        :param next_token: Used for pagination.
        :returns: GetPackageVersionHistoryResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("GetUpgradeHistory")
    def get_upgrade_history(
        self,
        context: RequestContext,
        domain_name: DomainName,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> GetUpgradeHistoryResponse:
        """Retrieves the complete history of the last 10 upgrades performed on the
        domain.

        :param domain_name: The name of an domain.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: Paginated APIs accept the NextToken input to return the next page of
        results and provide a NextToken output in the response, which you can
        use to retrieve more results.
        :returns: GetUpgradeHistoryResponse
        :raises BaseException:
        :raises ResourceNotFoundException:
        :raises DisabledOperationException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("GetUpgradeStatus")
    def get_upgrade_status(
        self, context: RequestContext, domain_name: DomainName
    ) -> GetUpgradeStatusResponse:
        """Retrieves the latest status of the last upgrade or upgrade eligibility
        check performed on the domain.

        :param domain_name: The name of an domain.
        :returns: GetUpgradeStatusResponse
        :raises BaseException:
        :raises ResourceNotFoundException:
        :raises DisabledOperationException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("ListDomainNames")
    def list_domain_names(
        self, context: RequestContext, engine_type: EngineType = None
    ) -> ListDomainNamesResponse:
        """Returns the names of all domains owned by the current user's account.

        :param engine_type: Optional parameter to filter the output by domain engine type.
        :returns: ListDomainNamesResponse
        :raises BaseException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("ListDomainsForPackage")
    def list_domains_for_package(
        self,
        context: RequestContext,
        package_id: PackageID,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> ListDomainsForPackageResponse:
        """Lists all Amazon OpenSearch Service domains associated with the package.

        :param package_id: The package for which to list associated domains.
        :param max_results: Limits the results to a maximum number of domains.
        :param next_token: Used for pagination.
        :returns: ListDomainsForPackageResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("ListInstanceTypeDetails")
    def list_instance_type_details(
        self,
        context: RequestContext,
        engine_version: VersionString,
        domain_name: DomainName = None,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> ListInstanceTypeDetailsResponse:
        """

        :param engine_version: .
        :param domain_name: The name of an domain.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: Paginated APIs accept the NextToken input to return the next page of
        results and provide a NextToken output in the response, which you can
        use to retrieve more results.
        :returns: ListInstanceTypeDetailsResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("ListPackagesForDomain")
    def list_packages_for_domain(
        self,
        context: RequestContext,
        domain_name: DomainName,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> ListPackagesForDomainResponse:
        """Lists all packages associated with the Amazon OpenSearch Service domain.

        :param domain_name: The name of the domain for which you want to list associated packages.
        :param max_results: Limits results to a maximum number of packages.
        :param next_token: Used for pagination.
        :returns: ListPackagesForDomainResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("ListTags")
    def list_tags(self, context: RequestContext, arn: ARN) -> ListTagsResponse:
        """Returns all tags for the given domain.

        :param arn: Specify the ``ARN`` of the domain that the tags you want to view are
        attached to.
        :returns: ListTagsResponse
        :raises BaseException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("ListVersions")
    def list_versions(
        self,
        context: RequestContext,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> ListVersionsResponse:
        """List all supported versions of OpenSearch and Elasticsearch.

        :param max_results: Set this value to limit the number of results returned.
        :param next_token: Paginated APIs accept the NextToken input to return the next page of
        results and provide a NextToken output in the response, which you can
        use to retrieve more results.
        :returns: ListVersionsResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("PurchaseReservedInstanceOffering")
    def purchase_reserved_instance_offering(
        self,
        context: RequestContext,
        reserved_instance_offering_id: GUID,
        reservation_name: ReservationToken,
        instance_count: InstanceCount = None,
    ) -> PurchaseReservedInstanceOfferingResponse:
        """Allows you to purchase reserved OpenSearch instances.

        :param reserved_instance_offering_id: The ID of the reserved OpenSearch instance offering to purchase.
        :param reservation_name: A customer-specified identifier to track this reservation.
        :param instance_count: The number of OpenSearch instances to reserve.
        :returns: PurchaseReservedInstanceOfferingResponse
        :raises ResourceNotFoundException:
        :raises ResourceAlreadyExistsException:
        :raises LimitExceededException:
        :raises DisabledOperationException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("RejectInboundConnection")
    def reject_inbound_connection(
        self, context: RequestContext, connection_id: ConnectionId
    ) -> RejectInboundConnectionResponse:
        """Allows the remote domain owner to reject an inbound cross-cluster
        connection request.

        :param connection_id: The ID of the inbound connection to reject.
        :returns: RejectInboundConnectionResponse
        :raises ResourceNotFoundException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("RemoveTags")
    def remove_tags(self, context: RequestContext, arn: ARN, tag_keys: StringList) -> None:
        """Removes the specified set of tags from the given domain.

        :param arn: The ``ARN`` of the domain from which you want to delete the specified
        tags.
        :param tag_keys: The ``TagKey`` list you want to remove from the domain.
        :raises BaseException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("StartServiceSoftwareUpdate")
    def start_service_software_update(
        self, context: RequestContext, domain_name: DomainName
    ) -> StartServiceSoftwareUpdateResponse:
        """Schedules a service software update for an Amazon OpenSearch Service
        domain.

        :param domain_name: The name of the domain that you want to update to the latest service
        software.
        :returns: StartServiceSoftwareUpdateResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("UpdateDomainConfig")
    def update_domain_config(
        self,
        context: RequestContext,
        domain_name: DomainName,
        cluster_config: ClusterConfig = None,
        ebs_options: EBSOptions = None,
        snapshot_options: SnapshotOptions = None,
        vpc_options: VPCOptions = None,
        cognito_options: CognitoOptions = None,
        advanced_options: AdvancedOptions = None,
        access_policies: PolicyDocument = None,
        log_publishing_options: LogPublishingOptions = None,
        encryption_at_rest_options: EncryptionAtRestOptions = None,
        domain_endpoint_options: DomainEndpointOptions = None,
        node_to_node_encryption_options: NodeToNodeEncryptionOptions = None,
        advanced_security_options: AdvancedSecurityOptionsInput = None,
        auto_tune_options: AutoTuneOptions = None,
        dry_run: DryRun = None,
    ) -> UpdateDomainConfigResponse:
        """Modifies the cluster configuration of the specified domain, such as
        setting the instance type and the number of instances.

        :param domain_name: The name of the domain you're updating.
        :param cluster_config: The type and number of instances to instantiate for the domain cluster.
        :param ebs_options: Specify the type and size of the EBS volume to use.
        :param snapshot_options: Option to set the time, in UTC format, for the daily automated snapshot.
        :param vpc_options: Options to specify the subnets and security groups for the VPC endpoint.
        :param cognito_options: Options to specify the Cognito user and identity pools for OpenSearch
        Dashboards authentication.
        :param advanced_options: Modifies the advanced option to allow references to indices in an HTTP
        request body.
        :param access_policies: IAM access policy as a JSON-formatted string.
        :param log_publishing_options: Map of ``LogType`` and ``LogPublishingOption``, each containing options
        to publish a given type of OpenSearch log.
        :param encryption_at_rest_options: Specifies encryption of data at rest options.
        :param domain_endpoint_options: Options to specify configuration that will be applied to the domain
        endpoint.
        :param node_to_node_encryption_options: Specifies node-to-node encryption options.
        :param advanced_security_options: Specifies advanced security options.
        :param auto_tune_options: Specifies Auto-Tune options.
        :param dry_run: This flag, when set to True, specifies whether the ``UpdateDomain``
        request should return the results of validation checks (DryRunResults)
        without actually applying the change.
        :returns: UpdateDomainConfigResponse
        :raises BaseException:
        :raises InternalException:
        :raises InvalidTypeException:
        :raises LimitExceededException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("UpdatePackage")
    def update_package(
        self,
        context: RequestContext,
        package_id: PackageID,
        package_source: PackageSource,
        package_description: PackageDescription = None,
        commit_message: CommitMessage = None,
    ) -> UpdatePackageResponse:
        """Updates a package for use with Amazon OpenSearch Service domains.

        :param package_id: The unique identifier for the package.
        :param package_source: The Amazon S3 location for importing the package specified as
        ``S3BucketName`` and ``S3Key``.
        :param package_description: A new description of the package.
        :param commit_message: A commit message for the new version which is shown as part of
        ``GetPackageVersionHistoryResponse``.
        :returns: UpdatePackageResponse
        :raises BaseException:
        :raises InternalException:
        :raises LimitExceededException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("UpgradeDomain")
    def upgrade_domain(
        self,
        context: RequestContext,
        domain_name: DomainName,
        target_version: VersionString,
        perform_check_only: Boolean = None,
        advanced_options: AdvancedOptions = None,
    ) -> UpgradeDomainResponse:
        """Allows you to either upgrade your domain or perform an upgrade
        eligibility check to a compatible version of OpenSearch or
        Elasticsearch.

        :param domain_name: The name of an domain.
        :param target_version: The version of OpenSearch you intend to upgrade the domain to.
        :param perform_check_only: When true, indicates that an upgrade eligibility check needs to be
        performed.
        :param advanced_options: Exposes select native OpenSearch configuration values from
        ``opensearch.
        :returns: UpgradeDomainResponse
        :raises BaseException:
        :raises ResourceNotFoundException:
        :raises ResourceAlreadyExistsException:
        :raises DisabledOperationException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError



def rename_params(func, rename_map):
    def do_rename(params, **kwargs):
        values = func(params, **kwargs) if func else params
        for old_param, new_param in rename_map.items():
            values[new_param] = values.pop(old_param, None)
        return values

    return do_rename



import base64
import copy
import hashlib
import heapq
import inspect
import json
import logging
import random
import re
import string
import threading
import time
from queue import Empty, PriorityQueue
from typing import Dict, List, NamedTuple, Optional, Set

from moto.sqs.models import BINARY_TYPE_FIELD_INDEX, STRING_TYPE_FIELD_INDEX
from moto.sqs.models import Message as MotoMessage

from localstack import config
from localstack.aws.api import CommonServiceException, RequestContext
from localstack.aws.api.sqs import (
    ActionNameList,
    AttributeNameList,
    AWSAccountIdList,
    BatchEntryIdsNotDistinct,
    BatchResultErrorEntry,
    BoxedInteger,
    ChangeMessageVisibilityBatchRequestEntryList,
    ChangeMessageVisibilityBatchResult,
    CreateQueueResult,
    DeleteMessageBatchRequestEntryList,
    DeleteMessageBatchResult,
    DeleteMessageBatchResultEntry,
    EmptyBatchRequest,
    GetQueueAttributesResult,
    GetQueueUrlResult,
    Integer,
    InvalidAttributeName,
    InvalidMessageContents,
    ListQueuesResult,
    ListQueueTagsResult,
    Message,
    MessageAttributeNameList,
    MessageBodyAttributeMap,
    MessageBodySystemAttributeMap,
    MessageNotInflight,
    MessageSystemAttributeName,
    PurgeQueueInProgress,
    QueueAttributeMap,
    QueueAttributeName,
    QueueDoesNotExist,
    QueueNameExists,
    ReceiptHandleIsInvalid,
    ReceiveMessageResult,
    SendMessageBatchRequestEntryList,
    SendMessageBatchResult,
    SendMessageBatchResultEntry,
    SendMessageResult,
    SqsApi,
    String,
    TagKeyList,
    TagMap,
    Token,
)
from localstack.aws.spec import load_service
from localstack.config import external_service_url
from localstack.services.plugins import ServiceLifecycleHook
from localstack.utils.aws.aws_stack import parse_arn
from localstack.utils.common import long_uid, md5, now, start_thread
from localstack.utils.run import FuncThread

LOG = logging.getLogger(__name__)

# Valid unicode values: #x9 | #xA | #xD | #x20 to #xD7FF | #xE000 to #xFFFD | #x10000 to #x10FFFF
# https://docs.aws.amazon.com/AWSSimpleQueueService/latest/APIReference/API_SendMessage.html
MSG_CONTENT_REGEX = "^[\u0009\u000A\u000D\u0020-\uD7FF\uE000-\uFFFD\U00010000-\U0010FFFF]*$"

# https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-message-metadata.html
# While not documented, umlauts seem to be allowed
ATTR_NAME_CHAR_REGEX = "^[\u00C0-\u017Fa-zA-Z0-9_.-]*$"
ATTR_NAME_PREFIX_SUFFIX_REGEX = r"^(?!(aws\.|amazon\.|\.)).*(?<!\.)$"
ATTR_TYPE_REGEX = "^(String|Number|Binary).*$"
FIFO_MSG_REGEX = "^[0-9a-zA-z!\"#$%&'()*+,./:;<=>?@[\\]^_`{|}~-]*$"

DEDUPLICATION_INTERVAL_IN_SEC = 5 * 60


def generate_message_id():
    return long_uid()


def generate_receipt_handle():
    # http://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/ImportantIdentifiers.html#ImportantIdentifiers-receipt-handles
    return "".join(random.choices(string.ascii_letters + string.digits, k=172)) + "="


class InvalidParameterValue(CommonServiceException):
    def __init__(self, message):
        super().__init__("InvalidParameterValues", message, 400, True)


class NonExistentQueue(CommonServiceException):
    def __init__(self):
        # TODO: not sure if this is really how AWS behaves
        super().__init__(
            "AWS.SimpleQueueService.NonExistentQueue",
            "The specified queue does not exist for this wsdl version.",
            status_code=400,
        )


class InvalidAttributeValue(CommonServiceException):
    def __init__(self, message):
        super().__init__("InvalidAttributeValue", message, 400, True)


class MissingParameter(CommonServiceException):
    def __init__(self, message):
        super().__init__("MissingParameter", message, 400, True)


def assert_queue_name(queue_name: str, fifo: bool = False):
    if queue_name.endswith(".fifo"):
        if not fifo:
            # Standard queues with .fifo suffix are not allowed
            raise InvalidParameterValue(
                "Can only include alphanumeric characters, hyphens, or underscores. 1 to 80 in length"
            )
        # The .fifo suffix counts towards the 80-character queue name quota.
        queue_name = queue_name[:-5] + "_fifo"

    # slashes are actually not allowed, but we've allowed it explicitly in localstack
    if not re.match(r"^[a-zA-Z0-9/_-]{1,80}$", queue_name):
        raise InvalidParameterValue(
            "Can only include alphanumeric characters, hyphens, or underscores. 1 to 80 in length"
        )


def check_message_content(message_body: str):
    error = "Invalid characters found. Valid unicode characters are #x9 | #xA | #xD | #x20 to #xD7FF | #xE000 to #xFFFD | #x10000 to #x10FFFF"

    if not re.match(MSG_CONTENT_REGEX, message_body):
        raise InvalidMessageContents(error)


class QueueKey(NamedTuple):
    region: str
    account_id: str
    name: str


class Permission(NamedTuple):
    # TODO: just a placeholder for real policies
    label: str
    account_id: str
    action: str


class SqsMessage:
    message: Message
    visibility_timeout: int
    receive_times: int
    receipt_handles: Set[str]
    deleted: bool
    priority: float
    message_deduplication_id: str
    message_group_id: str

    def __init__(
        self,
        priority: float,
        message: Message,
        message_deduplication_id: str = None,
        message_group_id: str = None,
    ) -> None:
        self.message = message
        self.receive_times = 0
        self.receipt_handles = set()

        self.last_received = None
        self.first_received = None
        self.deleted = False
        self.priority = priority
        self.message_deduplication_id = message_deduplication_id
        self.message_group_id = message_group_id

    @property
    def is_visible(self):
        if self.last_received is None:
            return True
        if time.time() >= (self.last_received + self.visibility_timeout):
            return True

        return False

    def __gt__(self, other):
        return self.priority > other.priority

    def __ge__(self, other):
        return self.priority >= other.priority

    def __lt__(self, other):
        return self.priority < other.priority

    def __le__(self, other):
        return self.priority <= other.priority

    def __eq__(self, other):
        return self.message["MessageId"] == other.message["MessageId"]

    def __hash__(self):
        return self.message["MessageId"].__hash__()


class SqsQueue:
    key: QueueKey

    attributes: QueueAttributeMap
    tags: TagMap
    permissions: Set[Permission]

    purge_in_progress: bool

    visible: PriorityQueue
    inflight: Set[SqsMessage]
    receipts: Dict[str, SqsMessage]

    def __init__(self, key: QueueKey, attributes=None, tags=None) -> None:
        super().__init__()
        self._assert_queue_name(key.name)
        self.key = key
        self.tags = tags or {}

        self.visible = PriorityQueue()
        self.inflight = set()
        self.receipts = {}

        self.attributes = self.default_attributes()
        if attributes:
            self.attributes.update(attributes)

        self.purge_in_progress = False
        self.permissions = set()
        self.mutex = threading.RLock()

    def default_attributes(self) -> QueueAttributeMap:
        return {
            QueueAttributeName.QueueArn: self.arn,
            QueueAttributeName.ApproximateNumberOfMessages: self.visible._qsize,
            QueueAttributeName.ApproximateNumberOfMessagesNotVisible: lambda: len(self.inflight),
            QueueAttributeName.ApproximateNumberOfMessagesDelayed: "0",  # FIXME: this should also be callable
            QueueAttributeName.CreatedTimestamp: str(now()),
            QueueAttributeName.LastModifiedTimestamp: str(now()),
            QueueAttributeName.VisibilityTimeout: "30",
            QueueAttributeName.MaximumMessageSize: "262144",
            QueueAttributeName.MessageRetentionPeriod: "345600",
            QueueAttributeName.DelaySeconds: "0",
            QueueAttributeName.ReceiveMessageWaitTimeSeconds: "0",
        }

    def update_last_modified(self, timestamp: int = None):
        if timestamp is None:
            timestamp = now()

        self.attributes[QueueAttributeName.LastModifiedTimestamp] = str(timestamp)

    @property
    def name(self):
        return self.key.name

    @property
    def owner(self):
        return self.key.account_id

    @property
    def arn(self) -> str:
        return f"arn:aws:sqs:{self.key.region}:{self.key.account_id}:{self.key.name}"

    def url(self, context: RequestContext) -> str:
        """Return queue URL using either SQS_PORT_EXTERNAL (if configured), or based on the 'Host' request header"""
        host_url = context.request.host_url
        if config.SQS_PORT_EXTERNAL:
            host_url = external_service_url("sqs")
        return "{host}/{account_id}/{name}".format(
            host=host_url.rstrip("/"),
            account_id=self.key.account_id,
            name=self.key.name,
        )

    @property
    def visibility_timeout(self) -> int:
        return int(self.attributes[QueueAttributeName.VisibilityTimeout])

    def update_visibility_timeout(self, receipt_handle: str, visibility_timeout: int):
        with self.mutex:
            if receipt_handle not in self.receipts:
                raise ReceiptHandleIsInvalid(
                    f'The input receipt handle "{receipt_handle}" is not a valid receipt handle.'
                )
            standard_message = self.receipts[receipt_handle]

            if standard_message not in self.inflight:
                raise MessageNotInflight()

            standard_message.visibility_timeout = visibility_timeout

            if visibility_timeout == 0:
                LOG.info(
                    "terminating the visibility timeout of %s",
                    standard_message.message["MessageId"],
                )
                # Terminating the visibility timeout for a message
                # https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-visibility-timeout.html#terminating-message-visibility-timeout
                self.inflight.remove(standard_message)
                self.visible.put_nowait(standard_message)

    def remove(self, receipt_handle: str):
        with self.mutex:
            if receipt_handle not in self.receipts:
                LOG.debug(
                    "no in-flight message found for receipt handle %s in queue %s",
                    receipt_handle,
                    self.arn,
                )
                return

            standard_message = self.receipts[receipt_handle]
            standard_message.deleted = True
            LOG.debug(
                "deleting message %s from queue %s", standard_message.message["MessageId"], self.arn
            )

            # remove all handles
            for handle in standard_message.receipt_handles:
                del self.receipts[handle]
            standard_message.receipt_handles.clear()

            # remove in-flight message
            try:
                self.inflight.remove(standard_message)
            except KeyError:
                # this means the message was re-queued in the meantime
                # TODO: remove this message from the visible queue if it exists: a message can be removed with an old
                #  receipt handle that was issued before the message was put back in the visible queue.
                self.visible.queue.remove(standard_message)
                heapq.heapify(self.visible.queue)
                pass

    def put(
        self,
        message: Message,
        visibility_timeout: int = None,
        message_deduplication_id: str = None,
        message_group_id: str = None,
    ):
        raise NotImplementedError

    def get(self, block=True, timeout=None, visibility_timeout: int = None) -> SqsMessage:
        start = time.time()
        while True:
            standard_message: SqsMessage = self.visible.get(block=block, timeout=timeout)
            LOG.debug(
                "de-queued message %s from %s", standard_message.message["MessageId"], self.arn
            )

            with self.mutex:
                if standard_message.deleted:
                    # TODO: check what the behavior of AWS is here. should we return a deleted message?
                    timeout -= time.time() - start
                    if timeout < 0:
                        timeout = 0
                    continue

                # update message attributes
                standard_message.visibility_timeout = (
                    self.visibility_timeout if visibility_timeout is None else visibility_timeout
                )
                standard_message.receive_times += 1
                standard_message.last_received = time.time()
                if standard_message.first_received is None:
                    standard_message.first_received = standard_message.last_received

                # create and manage receipt handle
                receipt_handle = generate_receipt_handle()
                standard_message.receipt_handles.add(receipt_handle)
                self.receipts[receipt_handle] = standard_message

                if standard_message.visibility_timeout == 0:
                    self.visible.put_nowait(standard_message)
                else:
                    self.inflight.add(standard_message)

                # prepare message for receiver
                # TODO: update message attributes (ApproximateFirstReceiveTimestamp, ApproximateReceiveCount)
                copied_message = copy.deepcopy(standard_message)
                copied_message.message["ReceiptHandle"] = receipt_handle

            return copied_message

    def requeue_inflight_messages(self):
        if not self.inflight:
            return

        with self.mutex:
            messages = list(self.inflight)
            for standard_message in messages:
                if standard_message.is_visible:
                    LOG.debug(
                        "re-queueing inflight messages %s into queue %s",
                        standard_message.message["MessageId"],
                        self.arn,
                    )
                    self.inflight.remove(standard_message)
                    self.visible.put_nowait(standard_message)

    def _assert_queue_name(self, name):
        if not re.match(r"^[a-zA-Z0-9_-]{1,80}$", name):
            raise InvalidParameterValue(
                "Can only include alphanumeric characters, hyphens, or underscores. 1 to 80 in length"
            )

    def validate_queue_attributes(self, attributes):
        valid = [k[1] for k in inspect.getmembers(QueueAttributeName)]
        del valid[valid.index(QueueAttributeName.FifoQueue)]

        for k in attributes.keys():
            if k not in valid:
                raise InvalidAttributeName(f"Unknown attribute name {k}")

    def generate_sequence_number(self):
        return None


class StandardQueue(SqsQueue):
    def put(
        self,
        message: Message,
        visibility_timeout: int = None,
        message_deduplication_id: str = None,
        message_group_id: str = None,
    ):

        if message_deduplication_id:
            raise InvalidParameterValue(
                f"Value {message_deduplication_id} for parameter MessageDeduplicationId is invalid. Reason: The "
                f"request includes a parameter that is not valid for this queue type. "
            )
        if message_group_id:
            raise InvalidParameterValue(
                f"Value {message_group_id} for parameter MessageGroupId is invalid. Reason: The request includes a "
                f"parameter that is not valid for this queue type. "
            )

        standard_message = SqsMessage(time.time(), message)

        if visibility_timeout is not None:
            standard_message.visibility_timeout = visibility_timeout
        else:
            # use the attribute from the queue
            standard_message.visibility_timeout = self.visibility_timeout

        self.visible.put_nowait(standard_message)


class FifoQueue(SqsQueue):
    visible: PriorityQueue
    inflight: Set[SqsMessage]
    receipts: Dict[str, SqsMessage]
    deduplication: Dict[str, Dict[str, SqsMessage]]

    def __init__(self, key: QueueKey, attributes=None, tags=None) -> None:
        super().__init__(key, attributes, tags)
        self.deduplication = {}

    def put(
        self,
        message: Message,
        visibility_timeout: int = None,
        message_deduplication_id: str = None,
        message_group_id: str = None,
    ):

        if not message_group_id:
            raise MissingParameter("The request must contain the parameter MessageGroupId.")
        dedup_id = message_deduplication_id
        content_based_deduplication = (
            "true"
            == (self.attributes.get(QueueAttributeName.ContentBasedDeduplication, "false")).lower()
        )
        if not dedup_id and content_based_deduplication:
            dedup_id = hashlib.sha256(message.get("Body").encode("utf-8")).hexdigest()
        if not dedup_id:
            raise InvalidParameterValue(
                "The Queue should either have ContentBasedDeduplication enabled or MessageDeduplicationId provided "
                "explicitly "
            )

        qm = SqsMessage(
            time.time(),
            message,
            message_deduplication_id=dedup_id,
            message_group_id=message_group_id,
        )
        if visibility_timeout is not None:
            qm.visibility_timeout = visibility_timeout
        else:
            # use the attribute from the queue
            qm.visibility_timeout = self.visibility_timeout
        original_message = None
        original_message_group = self.deduplication.get(message_group_id)
        if original_message_group:
            original_message = original_message_group.get(dedup_id)
        if (
            original_message
            and not original_message.deleted
            and original_message.priority + DEDUPLICATION_INTERVAL_IN_SEC > qm.priority
        ):
            message["MessageId"] = original_message.message["MessageId"]
        else:
            self.visible.put_nowait(qm)
            if not original_message_group:
                self.deduplication[message_group_id] = {}
            self.deduplication[message_group_id][message_deduplication_id] = qm

    def _assert_queue_name(self, name):
        if not name.endswith(".fifo"):
            raise InvalidParameterValue(
                "The name of a FIFO queue can only include alphanumeric characters, hyphens, or underscores, "
                "must end with .fifo suffix and be 1 to 80 in length"
            )
        # The .fifo suffix counts towards the 80-character queue name quota.
        queue_name = name[:-5] + "_fifo"
        super()._assert_queue_name(queue_name)

    def validate_queue_attributes(self, attributes):
        valid = [k[1] for k in inspect.getmembers(QueueAttributeName)]

        for k in attributes.keys():
            if k not in valid:
                raise InvalidAttributeName(f"Unknown attribute name {k}")
        # Special Cases
        fifo = attributes.get(QueueAttributeName.FifoQueue)
        if fifo and fifo.lower() != "true":
            raise InvalidAttributeValue(
                "Invalid value for the parameter FifoQueue. Reason: Modifying queue type is not supported."
            )

    # TODO: If we ever actually need to do something with this number, it needs to be part of
    #   SQSMessage. This means changing all *put*() signatures to return the saved message.
    def generate_sequence_number(self):
        return _create_mock_sequence_number()


class InflightUpdateWorker:
    """
    Regularly re-queues inflight messages whose visibility timeout has expired.

    FIXME: very crude implementation. it would be better to have event-driven communication.
    """

    queues: Dict[QueueKey, SqsQueue]

    def __init__(self, queues: Dict[QueueKey, SqsQueue]) -> None:
        super().__init__()
        self.queues = queues
        self.running = False
        self.thread: Optional[FuncThread] = None

    def start(self):
        if self.thread:
            return

        def _run(*_args):
            self.running = True
            self.run()

        self.thread = start_thread(_run)

    def stop(self):
        if self.thread:
            self.thread.stop()

        self.running = False
        self.thread = None

    def run(self):
        while self.running:
            time.sleep(1)
            for queue in self.queues.values():
                queue.requeue_inflight_messages()


def check_attributes(message_attributes: MessageBodyAttributeMap):
    if not message_attributes:
        return
    for attribute_name in message_attributes:
        if len(attribute_name) >= 256:
            raise InvalidParameterValue(
                "Message (user) attribute names must be shorter than 256 Bytes"
            )
        if not re.match(ATTR_NAME_CHAR_REGEX, attribute_name.lower()):
            raise InvalidParameterValue(
                "Message (user) attributes name can only contain upper and lower score characters, digits, periods, "
                "hyphens and underscores. "
            )
        if not re.match(ATTR_NAME_PREFIX_SUFFIX_REGEX, attribute_name.lower()):
            raise InvalidParameterValue(
                "You can't use message attribute names beginning with 'AWS.' or 'Amazon.'. "
                "These strings are reserved for internal use. Additionally, they cannot start or end with '.'."
            )

        attribute = message_attributes[attribute_name]
        attribute_type = attribute.get("DataType")
        if not attribute_type:
            raise InvalidParameterValue("Missing required parameter DataType")
        if not re.match(ATTR_TYPE_REGEX, attribute_type):
            raise InvalidParameterValue(
                f"Type for parameter MessageAttributes.Attribute_name.DataType must be prefixed"
                f'with "String", "Binary", or "Number", but was: {attribute_type}'
            )
        if len(attribute_type) >= 256:
            raise InvalidParameterValue(
                "Message (user) attribute types must be shorter than 256 Bytes"
            )

        if attribute_type == "String":
            try:
                attribute_value = attribute.get("StringValue")
                check_message_content(attribute_value)
            except InvalidMessageContents as e:
                # AWS throws a different exception here
                raise InvalidParameterValue(e.args[0])


def check_fifo_id(fifo_id):
    if not fifo_id:
        return
    if len(fifo_id) >= 128:
        raise InvalidParameterValue(
            "Message deduplication ID and group ID must be shorter than 128 bytes"
        )
    if not re.match(FIFO_MSG_REGEX, fifo_id):
        raise InvalidParameterValue(
            "Invalid characters found. Deduplication ID and group ID can only contain"
            "alphanumeric characters as well as TODO"
        )


class SqsProvider(SqsApi, ServiceLifecycleHook):
    """
    LocalStack SQS Provider.

    LIMITATIONS:
        - Pagination of results (NextToken)
        - Delivery guarantees
        - The region is not encoded in the queue URL
    """

    queues: Dict[QueueKey, SqsQueue]

    def __init__(self) -> None:
        super().__init__()
        self.queues = {}
        self._mutex = threading.RLock()
        self._inflight_worker = InflightUpdateWorker(self.queues)

    def start(self):
        self._inflight_worker.start()

    def shutdown(self):
        self._inflight_worker.stop()

    def on_before_start(self):
        self.start()

    def on_before_stop(self):
        self.shutdown()

    def _add_queue(self, queue: SqsQueue):
        with self._mutex:
            self.queues[queue.key] = queue

    def _require_queue(self, key: QueueKey) -> SqsQueue:
        """
        Returns the queue for the given key, or raises NonExistentQueue if it does not exist.

        :param key: the QueueKey to look for
        :returns: the queue
        :raises NonExistentQueue: if the queue does not exist
        """
        with self._mutex:
            if key not in self.queues:
                raise NonExistentQueue()

            return self.queues[key]

    def _require_queue_by_arn(self, queue_arn: str) -> SqsQueue:
        arn = parse_arn(queue_arn)
        key = QueueKey(region=arn["region"], account_id=arn["account"], name=arn["resource"])
        return self._require_queue(key)

    def _resolve_queue(
        self,
        context: RequestContext,
        queue_name: Optional[str] = None,
        queue_url: Optional[str] = None,
    ) -> SqsQueue:
        """
        Uses resolve_queue_key to determine the QueueKey from the given input, and returns the respective queue,
        or raises NonExistentQueue if it does not exist.

        :param context: the request context, used for getting region and account_id, and optionally the queue_url
        :param queue_name: the queue name (if this is set, then this will be used for the key)
        :param queue_url: the queue url (if name is not set, this will be used to determine the queue name)
        :returns: the queue
        :raises NonExistentQueue: if the queue does not exist
        """
        key = resolve_queue_key(context, queue_name, queue_url)
        return self._require_queue(key)

    def create_queue(
        self,
        context: RequestContext,
        queue_name: String,
        attributes: QueueAttributeMap = None,
        tags: TagMap = None,
    ) -> CreateQueueResult:
        fifo = attributes and (
            attributes.get(QueueAttributeName.FifoQueue, "false").lower() == "true"
        )

        k = QueueKey(context.region, context.account_id, queue_name)

        # Special Case TODO: why is an emtpy policy passed at all? same in set_queue_attributes
        if attributes and attributes.get(QueueAttributeName.Policy) == "":
            del attributes[QueueAttributeName.Policy]

        if k in self.queues:
            raise QueueNameExists(queue_name)
        if fifo:
            queue = FifoQueue(k, attributes, tags)
        else:
            queue = StandardQueue(k, attributes, tags)
        LOG.debug("creating queue key=%s attributes=%s tags=%s", k, attributes, tags)
        self._add_queue(queue)

        return CreateQueueResult(QueueUrl=queue.url(context))

    def get_queue_url(
        self, context: RequestContext, queue_name: String, queue_owner_aws_account_id: String = None
    ) -> GetQueueUrlResult:
        account_id = queue_owner_aws_account_id or context.account_id
        key = QueueKey(context.region, account_id, queue_name)

        if key not in self.queues:
            raise QueueDoesNotExist("The specified queue does not exist for this wsdl version.")

        queue = self.queues[key]
        self._assert_permission(context, queue)

        return GetQueueUrlResult(QueueUrl=queue.url(context))

    def list_queues(
        self,
        context: RequestContext,
        queue_name_prefix: String = None,
        next_token: Token = None,
        max_results: BoxedInteger = None,
    ) -> ListQueuesResult:
        urls = []

        for queue in self.queues.values():
            if queue.key.region != context.region:
                continue
            if queue.key.account_id != context.account_id:
                continue
            if queue_name_prefix:
                if not queue.name.startswith(queue_name_prefix):
                    continue
            urls.append(queue.url(context))

        if max_results:
            # FIXME: also need to solve pagination with stateful iterators: If the total number of items available is
            #  more than the value specified, a NextToken is provided in the command's output. To resume pagination,
            #  provide the NextToken value in the starting-token argument of a subsequent command. Do not use the
            #  NextToken response element directly outside of the AWS CLI.
            urls = urls[:max_results]

        return ListQueuesResult(QueueUrls=urls)

    def change_message_visibility(
        self,
        context: RequestContext,
        queue_url: String,
        receipt_handle: String,
        visibility_timeout: Integer,
    ) -> None:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)
        queue.update_visibility_timeout(receipt_handle, visibility_timeout)

    def change_message_visibility_batch(
        self,
        context: RequestContext,
        queue_url: String,
        entries: ChangeMessageVisibilityBatchRequestEntryList,
    ) -> ChangeMessageVisibilityBatchResult:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)

        self._assert_batch(entries)

        successful = []
        failed = []

        with queue.mutex:
            for entry in entries:
                try:
                    queue.update_visibility_timeout(
                        entry["ReceiptHandle"], entry["VisibilityTimeout"]
                    )
                    successful.append({"Id": entry["Id"]})
                except Exception as e:
                    failed.append(
                        BatchResultErrorEntry(
                            Id=entry["Id"],
                            SenderFault=False,
                            Code=e.__class__.__name__,
                            Message=str(e),
                        )
                    )

        return ChangeMessageVisibilityBatchResult(
            Successful=successful,
            Failed=failed,
        )

    def delete_queue(self, context: RequestContext, queue_url: String) -> None:
        with self._mutex:
            queue = self._resolve_queue(context, queue_url=queue_url)
            self._assert_permission(context, queue)
            del self.queues[queue.key]

    def get_queue_attributes(
        self, context: RequestContext, queue_url: String, attribute_names: AttributeNameList = None
    ) -> GetQueueAttributesResult:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)

        if not attribute_names:
            return GetQueueAttributesResult(Attributes={})

        if QueueAttributeName.All in attribute_names:
            # return GetQueueAttributesResult(Attributes=queue.attributes)
            attribute_names = queue.attributes.keys()

        result: Dict[QueueAttributeName, str] = {}

        for attr in attribute_names:
            try:
                getattr(QueueAttributeName, attr)
            except AttributeError:
                raise InvalidAttributeName(f"Unknown attribute {attr}.")

            if callable(queue.attributes.get(attr)):
                func = queue.attributes.get(attr)
                result[attr] = func()
            else:
                result[attr] = queue.attributes.get(attr)

        return GetQueueAttributesResult(Attributes=result)

    def send_message(
        self,
        context: RequestContext,
        queue_url: String,
        message_body: String,
        delay_seconds: Integer = None,
        message_attributes: MessageBodyAttributeMap = None,
        message_system_attributes: MessageBodySystemAttributeMap = None,
        message_deduplication_id: String = None,
        message_group_id: String = None,
    ) -> SendMessageResult:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)

        message = self._put_message(
            queue,
            context,
            message_body,
            delay_seconds,
            message_attributes,
            message_system_attributes,
            message_deduplication_id,
            message_group_id,
        )
        return SendMessageResult(
            MessageId=message["MessageId"],
            MD5OfMessageBody=message["MD5OfBody"],
            MD5OfMessageAttributes=message.get("MD5OfMessageAttributes"),
            SequenceNumber=queue.generate_sequence_number(),
            MD5OfMessageSystemAttributes=_create_message_attribute_hash(message_system_attributes),
        )

    def send_message_batch(
        self, context: RequestContext, queue_url: String, entries: SendMessageBatchRequestEntryList
    ) -> SendMessageBatchResult:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)

        self._assert_batch(entries)

        successful = []
        failed = []

        with queue.mutex:
            for entry in entries:
                try:
                    message = self._put_message(
                        queue,
                        context,
                        message_body=entry.get("MessageBody"),
                        delay_seconds=entry.get("DelaySeconds"),
                        message_attributes=entry.get("MessageAttributes"),
                        message_system_attributes=entry.get("MessageSystemAttributes"),
                        message_deduplication_id=entry.get("MessageDeduplicationId"),
                        message_group_id=entry.get("MessageGroupId"),
                    )

                    successful.append(
                        SendMessageBatchResultEntry(
                            Id=entry["Id"],
                            MessageId=message.get("MessageId"),
                            MD5OfMessageBody=message.get("MD5OfBody"),
                            MD5OfMessageAttributes=message.get("MD5OfMessageAttributes"),
                            MD5OfMessageSystemAttributes=_create_message_attribute_hash(
                                message.get("message_system_attributes")
                            ),
                            SequenceNumber=queue.generate_sequence_number(),
                        )
                    )
                except Exception as e:
                    failed.append(
                        BatchResultErrorEntry(
                            Id=entry["Id"],
                            SenderFault=False,
                            Code=e.__class__.__name__,
                            Message=str(e),
                        )
                    )

        return SendMessageBatchResult(
            Successful=successful,
            Failed=failed,
        )

    def _put_message(
        self,
        queue: SqsQueue,
        context: RequestContext,
        message_body: String,
        delay_seconds: Integer = None,
        message_attributes: MessageBodyAttributeMap = None,
        message_system_attributes: MessageBodySystemAttributeMap = None,
        message_deduplication_id: String = None,
        message_group_id: String = None,
    ) -> Message:
        # TODO: default message attributes (SenderId, ApproximateFirstReceiveTimestamp, ...)

        check_message_content(message_body)
        check_attributes(message_attributes)
        check_attributes(message_system_attributes)
        check_fifo_id(message_deduplication_id)
        check_fifo_id(message_group_id)

        message: Message = Message(
            MessageId=generate_message_id(),
            MD5OfBody=md5(message_body),
            Body=message_body,
            Attributes=self._create_message_attributes(context, message_system_attributes),
            MD5OfMessageAttributes=_create_message_attribute_hash(message_attributes),
            MessageAttributes=message_attributes,
        )
        delay_seconds = delay_seconds or queue.attributes.get(QueueAttributeName.DelaySeconds, "0")

        if int(delay_seconds):
            # FIXME: this is a pretty bad implementation (one thread per message...). polling on a priority queue
            #  would probably be better. We also need access to delayed messages for the
            #  ApproximateNumberrOfDelayedMessages attribute.
            threading.Timer(
                int(delay_seconds),
                queue.put,
                args=(message, message_deduplication_id, message_group_id),
            ).start()
        else:
            queue.put(
                message=message,
                message_deduplication_id=message_deduplication_id,
                message_group_id=message_group_id,
            )

        return message

    def receive_message(
        self,
        context: RequestContext,
        queue_url: String,
        attribute_names: AttributeNameList = None,
        message_attribute_names: MessageAttributeNameList = None,
        max_number_of_messages: Integer = None,
        visibility_timeout: Integer = None,
        wait_time_seconds: Integer = None,
        receive_request_attempt_id: String = None,
    ) -> ReceiveMessageResult:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)

        num = max_number_of_messages or 1
        block = wait_time_seconds is not None
        # collect messages
        messages = []
        while num:
            try:
                standard_message = queue.get(
                    block=block, timeout=wait_time_seconds, visibility_timeout=visibility_timeout
                )
                msg = standard_message.message
            except Empty:
                break

            moved_to_dlq = False
            if (
                queue.attributes
                and queue.attributes.get(QueueAttributeName.RedrivePolicy) is not None
            ):
                moved_to_dlq = self._dead_letter_check(queue, standard_message, context)
            if moved_to_dlq:
                continue
            # filter attributes
            if message_attribute_names:
                if "All" not in message_attribute_names:
                    msg["MessageAttributes"] = {
                        k: v
                        for k, v in msg["MessageAttributes"].items()
                        if k in message_attribute_names
                    }
                # TODO: why is this called even if we receive "All" attributes?
                msg["MD5OfMessageAttributes"] = _create_message_attribute_hash(
                    msg["MessageAttributes"]
                )
            else:
                del msg["MessageAttributes"]

            # add message to result
            messages.append(msg)
            num -= 1

        # TODO: how does receiving behave if the queue was deleted in the meantime?
        return ReceiveMessageResult(Messages=messages)

    def _dead_letter_check(
        self, queue: SqsQueue, std_m: SqsMessage, context: RequestContext
    ) -> bool:
        redrive_policy = json.loads(queue.attributes.get(QueueAttributeName.RedrivePolicy))
        # TODO: include the names of the dictionary sub - attributes in the autogenerated code?
        max_receive_count = redrive_policy["maxReceiveCount"]
        if std_m.receive_times > max_receive_count:
            dead_letter_target_arn = redrive_policy["deadLetterTargetArn"]
            dl_queue = self._require_queue_by_arn(dead_letter_target_arn)
            # TODO: this needs to be atomic?
            dead_message = std_m.message
            dl_queue.put(message=dead_message)
            queue.remove(std_m.message["ReceiptHandle"])
            return True
        else:
            return False

    def delete_message(
        self, context: RequestContext, queue_url: String, receipt_handle: String
    ) -> None:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)
        queue.remove(receipt_handle)

    def delete_message_batch(
        self,
        context: RequestContext,
        queue_url: String,
        entries: DeleteMessageBatchRequestEntryList,
    ) -> DeleteMessageBatchResult:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)
        self._assert_batch(entries)

        successful = []
        failed = []

        with queue.mutex:
            for entry in entries:
                try:
                    queue.remove(entry["ReceiptHandle"])
                    successful.append(DeleteMessageBatchResultEntry(Id=entry["Id"]))
                except Exception as e:
                    failed.append(
                        BatchResultErrorEntry(
                            Id=entry["Id"],
                            SenderFault=False,
                            Code=e.__class__.__name__,
                            Message=str(e),
                        )
                    )

        return DeleteMessageBatchResult(
            Successful=successful,
            Failed=failed,
        )

    def purge_queue(self, context: RequestContext, queue_url: String) -> None:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)

        with self._mutex:
            # FIXME: use queue-specific locks
            if queue.purge_in_progress:
                raise PurgeQueueInProgress()
            queue.purge_in_progress = True

        # TODO: how do other methods behave when purge is in progress?

        try:
            while True:
                queue.visible.get_nowait()
        except Empty:
            return
        finally:
            queue.purge_in_progress = False

    def set_queue_attributes(
        self, context: RequestContext, queue_url: String, attributes: QueueAttributeMap
    ) -> None:
        queue = self._resolve_queue(context, queue_url=queue_url)

        if not attributes:
            return

        queue.validate_queue_attributes(attributes)

        for k, v in attributes.items():
            queue.attributes[k] = v

        # Special cases
        if queue.attributes.get(QueueAttributeName.Policy) == "":
            del queue.attributes[QueueAttributeName.Policy]

        redrive_policy = queue.attributes.get(QueueAttributeName.RedrivePolicy)
        if redrive_policy:
            _redrive_policy = json.loads(redrive_policy)
            dl_target_arn = _redrive_policy.get("deadLetterTargetArn")
            max_receive_count = _redrive_policy.get("maxReceiveCount")
            # TODO: use the actual AWS responses
            if not dl_target_arn:
                raise InvalidParameterValue(
                    "The required parameter 'deadLetterTargetArn' is missing"
                )
            if not max_receive_count:
                raise InvalidParameterValue("The required parameter 'maxReceiveCount' is missing")
            try:
                max_receive_count = int(max_receive_count)
                valid_count = 1 <= max_receive_count <= 1000
            except ValueError:
                valid_count = False
            if not valid_count:
                raise InvalidParameterValue(
                    f"Value {redrive_policy} for parameter RedrivePolicy is invalid. Reason: Invalid value for maxReceiveCount: {max_receive_count}, valid values are from 1 to 1000 both inclusive."
                )

    def tag_queue(self, context: RequestContext, queue_url: String, tags: TagMap) -> None:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)

        if not tags:
            return

        for k, v in tags.items():
            queue.tags[k] = v

    def list_queue_tags(self, context: RequestContext, queue_url: String) -> ListQueueTagsResult:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)
        return ListQueueTagsResult(Tags=queue.tags)

    def untag_queue(self, context: RequestContext, queue_url: String, tag_keys: TagKeyList) -> None:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)

        for k in tag_keys:
            if k in queue.tags:
                del queue.tags[k]

    def add_permission(
        self,
        context: RequestContext,
        queue_url: String,
        label: String,
        aws_account_ids: AWSAccountIdList,
        actions: ActionNameList,
    ) -> None:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)

        self._validate_actions(actions)

        for account_id in aws_account_ids:
            for action in actions:
                queue.permissions.add(Permission(label, account_id, action))

    def remove_permission(self, context: RequestContext, queue_url: String, label: String) -> None:
        queue = self._resolve_queue(context, queue_url=queue_url)
        self._assert_permission(context, queue)

        candidates = [p for p in queue.permissions if p.label == label]
        if candidates:
            queue.permissions.remove(candidates[0])

    def _create_message_attributes(
        self,
        context: RequestContext,
        message_system_attributes: MessageBodySystemAttributeMap = None,
    ) -> Dict[MessageSystemAttributeName, str]:
        result: Dict[MessageSystemAttributeName, str] = {
            MessageSystemAttributeName.SenderId: context.account_id,
            MessageSystemAttributeName.SentTimestamp: str(now()),
        }

        if message_system_attributes is not None:
            for attr in message_system_attributes:
                result[attr] = message_system_attributes[attr]["StringValue"]

        return result

    def _validate_queue_attributes(self, attributes: QueueAttributeMap):
        valid = [k[1] for k in inspect.getmembers(QueueAttributeName)]

        for k in attributes.keys():
            if k not in valid:
                raise InvalidAttributeName("Unknown Attribute %s" % k)

    def _validate_actions(self, actions: ActionNameList):
        service = load_service(service=self.service, version=self.version)
        # FIXME: this is a bit of a heuristic as it will also include actions like "ListQueues" which is not
        #  associated with an action on a queue
        valid = list(service.operation_names)
        valid.append("*")

        for action in actions:
            if action not in valid:
                raise InvalidParameterValue(
                    f"Value SQS:{action} for parameter ActionName is invalid. Reason: Please refer to the appropriate "
                    "WSDL for a list of valid actions. "
                )

    def _assert_permission(self, context: RequestContext, queue: SqsQueue):
        action = context.operation.name
        account_id = context.account_id

        if account_id == queue.owner:
            return

        for permission in queue.permissions:
            if permission.account_id != account_id:
                continue
            if permission.action == "*":
                return
            if permission.action == action:
                return

        raise CommonServiceException("AccessDeniedException", "Not allowed")

    def _assert_batch(self, batch: List):
        if not batch:
            raise EmptyBatchRequest
        visited = set()
        for entry in batch:
            # TODO: InvalidBatchEntryId
            if entry["Id"] in visited:
                raise BatchEntryIdsNotDistinct()
            else:
                visited.add(entry["Id"])


def _create_mock_sequence_number():
    return "".join(random.choice(string.digits) for _ in range(20))


# Method from moto's attribute_md5 of moto/sqs/models.py, separated from the Message Object
def _create_message_attribute_hash(message_attributes) -> Optional[str]:
    # To avoid the need to check for dict conformity everytime we invoke this function
    if not isinstance(message_attributes, dict):
        return
    hash = hashlib.md5()

    for attrName in sorted(message_attributes.keys()):
        attr_value = message_attributes[attrName]
        # Encode name
        MotoMessage.update_binary_length_and_value(hash, MotoMessage.utf8(attrName))
        # Encode data type
        MotoMessage.update_binary_length_and_value(hash, MotoMessage.utf8(attr_value["DataType"]))
        # Encode transport type and value
        if attr_value.get("StringValue"):
            hash.update(bytearray([STRING_TYPE_FIELD_INDEX]))
            MotoMessage.update_binary_length_and_value(
                hash, MotoMessage.utf8(attr_value.get("StringValue"))
            )
        elif attr_value.get("BinaryValue"):
            hash.update(bytearray([BINARY_TYPE_FIELD_INDEX]))
            decoded_binary_value = base64.b64decode(attr_value.get("BinaryValue"))
            MotoMessage.update_binary_length_and_value(hash, decoded_binary_value)
        # string_list_value, binary_list_value type is not implemented, reserved for the future use.
        # See https://docs.aws.amazon.com/AWSSimpleQueueService/latest/APIReference/API_MessageAttributeValue.html
    return hash.hexdigest()


def get_queue_name_from_url(queue_url: str) -> str:
    return queue_url.rstrip("/").split("/")[-1]


def resolve_queue_key(
    context: RequestContext, queue_name: Optional[str] = None, queue_url: Optional[str] = None
) -> QueueKey:
    """
    Resolves a QueueKey from the given information.

    :param context: the request context, used for getting region and account_id, and optionally the queue_url
    :param queue_name: the queue name (if this is set, then this will be used for the key)
    :param queue_url: the queue url (if name is not set, this will be used to determine the queue name)
    :return: the QueueKey describing the queue being requested
    """
    if not queue_name:
        if queue_url:
            queue_name = get_queue_name_from_url(queue_url)
        else:
            queue_name = get_queue_name_from_url(context.request.base_url)

    return QueueKey(region=context.region, account_id=context.account_id, name=queue_name)



import dataclasses
import functools
import logging
import os
import platform

from localstack import config, constants
from localstack.runtime import hooks
from localstack.utils.common import FileMappedDocument, call_safe, long_uid, md5, short_uid

LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class ClientMetadata:
    session_id: str
    machine_id: str
    api_key: str
    system: str
    version: str
    is_ci: bool
    is_docker: bool
    is_testing: bool

    def __repr__(self):
        d = dataclasses.asdict(self)

        # anonymize api_key
        k = d.get("api_key")
        if k:
            k = "*" * len(k)
        d["api_key"] = k

        return "ClientMetadata(%s)" % d


def get_version_string() -> str:
    gh = config.LOCALSTACK_BUILD_GIT_HASH
    if gh:
        return f"{constants.VERSION}:{gh}"
    else:
        return constants.VERSION


def read_client_metadata() -> ClientMetadata:
    return ClientMetadata(
        session_id=get_session_id(),
        machine_id=get_machine_id(),
        api_key=read_api_key_safe(),
        system=get_system(),
        version=get_version_string(),
        is_ci=os.getenv("CI") is not None,
        is_docker=config.is_in_docker,
        is_testing=config.is_env_true(constants.ENV_INTERNAL_TEST_RUN),
    )


@functools.lru_cache()
def get_session_id() -> str:
    return _generate_session_id()


@functools.lru_cache()
def get_client_metadata() -> ClientMetadata:
    metadata = read_client_metadata()

    if config.DEBUG_ANALYTICS:
        LOG.info("resolved client metadata: %s", metadata)

    return metadata


@functools.lru_cache()
def get_machine_id() -> str:
    cache_path = os.path.join(config.dirs.cache, "machine.json")
    doc = FileMappedDocument(cache_path)

    if "machine_id" not in doc:
        # generate a machine id
        doc["machine_id"] = _generate_machine_id()
        # try to cache the machine ID
        call_safe(doc.save)

    return doc["machine_id"]


@hooks.prepare_host()
def prepare_host_machine_id():
    # lazy-init machine ID into cache on the host, which can then be used in the container
    get_machine_id()


def _generate_session_id() -> str:
    return long_uid()


def _generate_machine_id() -> str:
    if config.is_in_docker:
        return short_uid()

    # this can potentially be useful when generated on the host using the CLI and then mounted into the container via
    # machine.json
    try:
        if os.path.exists("/etc/machine-id"):
            with open("/etc/machine-id") as fd:
                return md5(str(fd.read()))[:8]
    except Exception:
        pass

    # always fall back to short_uid()
    return short_uid()


def read_api_key_safe():
    try:
        from localstack_ext.bootstrap.licensing import read_api_key

        return read_api_key(raise_if_missing=False)
    except Exception:
        return None


def get_system() -> str:
    return platform.system()



import json

from localstack.services.cloudformation.service_models import REF_ID_ATTRS, GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.common import short_uid


class KMSKey(GenericBaseModel):
    @staticmethod
    def cloudformation_type():
        return "AWS::KMS::Key"

    def fetch_state(self, stack_name, resources):
        client = aws_stack.connect_to_service("kms")
        physical_res_id = self.physical_resource_id
        props = self.props
        res_tags = props.get("Tags", [])
        if not physical_res_id:
            # TODO: find a more efficient approach for this?
            for key in client.list_keys()["Keys"]:
                details = client.describe_key(KeyId=key["KeyId"])["KeyMetadata"]
                tags = client.list_resource_tags(KeyId=key["KeyId"]).get("Tags", [])
                tags = [{"Key": tag["TagKey"], "Value": tag["TagValue"]} for tag in tags]
                if (
                    tags == res_tags
                    and details.get("Description") == props.get("Description")
                    and props.get("KeyUsage") in [None, details.get("KeyUsage")]
                ):
                    physical_res_id = key["KeyId"]
                    # TODO should this be removed from here? It seems that somewhere along the execution
                    #  chain the 'PhysicalResourceId' gets overwritten with None, hence setting it here
                    self.resource_json["PhysicalResourceId"] = physical_res_id
                    break
        if not physical_res_id:
            return
        return client.describe_key(KeyId=physical_res_id)

    def get_physical_resource_id(self, attribute=None, **kwargs):
        if attribute in REF_ID_ATTRS:
            return self.physical_resource_id
        return self.physical_resource_id and aws_stack.kms_key_arn(self.physical_resource_id)

    # TODO: try to remove this workaround (ensures idempotency)
    @staticmethod
    def add_defaults(resource, stack_name: str):
        props = resource["Properties"] = resource.get("Properties", {})
        tags = props["Tags"] = props.get("Tags", [])
        existing = [t for t in tags if t["Key"] == "localstack-key-id"]
        if not existing:
            # append tags, to allow us to determine in fetch_state whether this key is already deployed
            tags.append({"Key": "localstack-key-id", "Value": short_uid()})

    @classmethod
    def get_deploy_templates(cls):
        def _create(resource_id, resources, resource_type, func, stack_name):
            kms_client = aws_stack.connect_to_service("kms")
            resource = cls(resources[resource_id])
            props = resource.props
            params = {}
            if props.get("KeyPolicy"):
                params["Policy"] = json.dumps(props["KeyPolicy"])

            if props.get("Tags"):
                params["Tags"] = [
                    {"TagKey": tag["Key"], "TagValue": tag["Value"]}
                    for tag in props.get("Tags", [])
                ]

            if props.get("Description"):
                params["Description"] = props["Description"]

            if props.get("KeySpec"):
                params["KeySpec"] = props["KeySpec"]

            new_key = kms_client.create_key(**params)
            key_id = new_key["KeyMetadata"]["KeyId"]
            resource.resource_json["PhysicalResourceId"] = key_id

            # key is created but some fields map to separate api calls
            if props.get("EnableKeyRotation", False):
                kms_client.enable_key_rotation(KeyId=key_id)
            else:
                kms_client.disable_key_rotation(KeyId=key_id)

            if props.get("Enabled", True):
                kms_client.enable_key(KeyId=key_id)
            else:
                kms_client.disable_key(KeyId=key_id)

        return {
            "create": [
                {"function": _create},
            ],
            "delete": {
                # TODO Key needs to be deleted in KMS backend
                "function": "schedule_key_deletion",
                "parameters": {"KeyId": "PhysicalResourceId"},
            },
        }


class KMSAlias(GenericBaseModel):
    @staticmethod
    def cloudformation_type():
        return "AWS::KMS::Alias"

    def fetch_state(self, stack_name, resources):
        kms = aws_stack.connect_to_service("kms")
        aliases = kms.list_aliases()["Aliases"]
        for alias in aliases:
            if alias["AliasName"] == self.props.get("AliasName"):
                return alias

        return None

    @staticmethod
    def get_deploy_templates():
        return {
            "create": {
                "function": "create_alias",
                "parameters": {"AliasName": "AliasName", "TargetKeyId": "TargetKeyId"},
            },
            "delete": {
                "function": "delete_alias",
                "parameters": {"AliasName": "AliasName"},
            },
        }
